# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import yaml
import json
import os
import pathlib
import pickle
import time
from copy import deepcopy
from functools import wraps
from threading import Event
from typing import List, Any, Optional

from dateutil.tz import gettz
from json_database import JsonStorage
from ovos_bus_client import Message
from ovos_plugin_manager.language import OVOSLangDetectionFactory, OVOSLangTranslationFactory
from ovos_utils.gui import is_gui_connected
from ovos_utils.log import LOG, log_deprecation, deprecated
from ovos_utils.skills import get_non_properties
from ovos_utils.skills.settings import save_settings
from ovos_utils.xdg_utils import xdg_cache_home
from ovos_workshop.skills import OVOSSkill
from ovos_workshop.skills.fallback import FallbackSkillV1

from neon_utils.cache_utils import LRUCache
from neon_utils.file_utils import resolve_neon_resource_file
from neon_utils.location_utils import to_system_time
from neon_utils.message_utils import dig_for_message, resolve_message, get_message_user
from neon_utils.skills.neon_skill import CACHE_TIME_OFFSET, DEFAULT_SPEED_MODE, SPEED_MODE_EXTENSION_TIME
from neon_utils.user_utils import get_user_prefs


class NeonFallbackSkill(FallbackSkillV1):
    """
    Class that extends the NeonSkill and FallbackSkill classes to provide
    NeonSkill functionality to any Fallback skill subclassing this class.
    """
    def __init__(self, *args, **kwargs):
        log_deprecation("This class is deprecated. Implement "
                        "`ovos_workshop.skills.fallback.FallbackSkill`",
                        "2.0.0")

        FallbackSkillV1.__init__(self, *args, **kwargs)
        LOG.debug(f"instance_handlers={self.instance_fallback_handlers}")
        LOG.debug(f"class_handlers={FallbackSkillV1.fallback_handlers}")

        # Manual init of NeonSkill
        self.cache_loc = os.path.join(xdg_cache_home(), "neon")
        os.makedirs(self.cache_loc, exist_ok=True)
        self.lru_cache = LRUCache()
        self._gui_connected = False

        try:
            import neon_core
            self._neon_core = True
        except ImportError:
            self._neon_core = False

        self._actions_to_confirm = dict()

        self._lang_detector = None
        self._translator = None

        # TODO: Should below defaults be global config?
        # allow skills to specify timeout overrides per-skill
        self._speak_timeout = 30
        self._get_response_timeout = 15  # 10 for listener, 5 for STT, then timeout

    def initialize(self):
        # schedule an event to load the cache on disk every CACHE_TIME_OFFSET seconds
        self.schedule_event(self._write_cache_on_disk, CACHE_TIME_OFFSET,
                            name="neon.load_cache_on_disk")

    @property
    # @deprecated("Call `dateutil.tz.gettz` directly", "2.0.0")
    def sys_tz(self):
        # TODO: Is this deprecated?
        return gettz()

    @property
    @deprecated("Nothing should depend on `neon_core` vs other cores", "2.0.0")
    def neon_core(self):
        return self._neon_core

    @property
    @deprecated("Skills should track this internally or use converse",
                "2.0.0")
    def actions_to_confirm(self) -> dict:
        return self._actions_to_confirm

    @actions_to_confirm.setter
    @deprecated("Skills should track this internally or use converse",
                "2.0.0")
    def actions_to_confirm(self, val: dict):
        self._actions_to_confirm = val

    @property
    def lang_detector(self):
        if not self._lang_detector and OVOSLangDetectionFactory:
            try:
                self._lang_detector = \
                    OVOSLangDetectionFactory.create(self.config_core)
            except ValueError as x:
                LOG.error(f"Configured lang plugins not available: {x}")
        return self._lang_detector

    @property
    def translator(self):
        if not self._translator and OVOSLangTranslationFactory:
            try:
                self._translator = \
                    OVOSLangTranslationFactory.create(self.config_core)
            except ValueError as x:
                LOG.error(f"Configured lang plugins not available: {x}")
        return self._translator

    @property
    @deprecated("This is now configured in CommonQuery", "2.0.0")
    def skill_mode(self) -> str:
        """
        Determine the "speed mode" requested by the user
        """
        return get_user_prefs(dig_for_message()).get(
            'response_mode', {}).get('speed_mode') or DEFAULT_SPEED_MODE

    @property
    @deprecated("This is now configured in CommonQuery", "2.0.0")
    def extension_time(self) -> int:
        """
        Determine how long the skill should extend CommonSkill request timeouts
        """
        return SPEED_MODE_EXTENSION_TIME.get(self.skill_mode) or 10

    @property
    @deprecated("Always emit GUI events for the GUI module to manage", "2.0.0")
    def gui_enabled(self) -> bool:
        """
        If True, skill should display GUI pages
        """
        try:
            self._gui_connected = self._gui_connected or \
                                  is_gui_connected(self.bus)
            return self._gui_connected
        except Exception as x:
            # In container environments, this check fails so assume True
            LOG.exception(x)
            return True

    @property
    @deprecated("reference `self.settings` directly", "2.0.0")
    def ngi_settings(self):
        return self.preference_skill()

    @deprecated("reference `self.settings` directly", "2.0.0")
    def preference_skill(self, message=None) -> dict:
        """
        Returns the skill settings configuration
        Equivalent to self.settings if settings not in message context
        :param message: Message associated with request
        :return: dict of skill preferences
        """
        message = message or dig_for_message()
        return get_user_prefs(
            message).get("skills", {}).get(self.skill_id) or self.settings

    @deprecated("implement `neon_utils.user_utils.update_user_profile`",
                "2.0.0")
    def update_profile(self, new_preferences: dict, message: Message = None):
        """
        Updates a user profile with the passed new_preferences
        :param new_preferences: dict of updated preference values.
            Should follow {section: {key: val}} format
        :param message: Message associated with request
        """
        from neon_utils.user_utils import update_user_profile
        message = message or dig_for_message()
        try:
            update_user_profile(new_preferences, message, self.bus)
        except Exception as x:
            LOG.error(x)

    @resolve_message
    def update_skill_settings(self, new_preferences: dict,
                              message: Message = None, skill_global=True):
        """
        Updates skill settings with the passed new_preferences
        :param new_preferences: dict of updated preference values. {key: val}
        :param message: Message associated with request
        :param skill_global: Boolean to indicate these are
            global/non-user-specific variables
        """
        # TODO: Spec how to handle global vs per-user settings
        LOG.debug(f"Update skill settings with new: {new_preferences}")
        new_settings = {**self.preference_skill(message), **new_preferences}
        if not skill_global:
            new_preferences["skill_id"] = self.skill_id
            self.update_profile({"skills": {self.skill_id: new_settings}},
                                message)
        else:
            self.settings = new_settings
            if isinstance(self.settings, JsonStorage):
                self.settings.store()
            else:
                save_settings(self.file_system.path, self.settings)

    def send_with_audio(self, text_shout, audio_file, message, lang="en-us",
                        private=False, speaker=None):
        """
        Sends a Neon response with the passed text phrase and audio file
        :param text_shout: (str) Text to shout
        :param audio_file: (str) Full path to an arbitrary audio file to
            attach to shout; must be readable/accessible
        :param message: Message associated with request
        :param lang: (str) Language of wav_file
        :param private: (bool) Whether or not shout is private to the user
        :param speaker: (dict) Message sender data
        """
        # TODO: Update 'speak' to handle audio files
        # from shutil import copyfile
        if not speaker:
            speaker = {"name": "Neon", "language": None, "gender": None,
                       "voice": None}

        # Play this back regardless of user prefs
        speaker["override_user"] = True

        # Either gender should be fine
        responses = {lang: {"sentence": text_shout,
                            "male": audio_file,
                            "female": audio_file}}
        message.context["private"] = private
        LOG.info(f"sending klat.response with responses={responses} | "
                 f"speaker={speaker}")
        self.bus.emit(message.forward("klat.response",
                                      {"responses": responses,
                                       "speaker": speaker}))

    @deprecated("implement `neon_utils.user_utils.neon_must_respond`", "2.0.0")
    def neon_must_respond(self, message: Message = None) -> bool:
        """
        Checks if Neon must respond to an utterance (i.e. a server request)
        :param message: message associated with user request
        :returns: True if Neon must provide a response to this request
        """
        from neon_utils.message_utils import neon_must_respond
        return neon_must_respond(message)

    def voc_match(self, utt, voc_filename, lang=None, exact=False):
        # TODO: This should be addressed in vocab resolver classes
        try:
            super_return = super().voc_match(utt, voc_filename, lang, exact)
        except FileNotFoundError:
            super_return = None
        lang = lang or self.lang

        if super_return:
            return super_return
        elif super_return is False:
            if self.voc_match_cache.get(lang + voc_filename):
                return super_return

        LOG.warning(f"`{voc_filename}` not found, checking in neon_core")
        voc = resolve_neon_resource_file(f"text/{lang}/{voc_filename}.voc")
        if not voc:
            raise FileNotFoundError(voc)
        from ovos_utils.file_utils import read_vocab_file
        from itertools import chain
        import re
        vocab = read_vocab_file(voc)
        cache_key = lang + voc_filename
        self.voc_match_cache[cache_key] = list(chain(*vocab))
        if utt:
            if exact:
                # Check for exact match
                return any(i.strip() == utt
                           for i in self.voc_match_cache[cache_key])
            else:
                # Check for matches against complete words
                return any([re.match(r'.*\b' + i + r'\b.*', utt)
                            for i in self.voc_match_cache[cache_key]])
        else:
            return False

    @deprecated("WW status can be queried via messagebus", "2.0.0")
    def neon_in_request(self, message: Message) -> bool:
        """
        Checks if the utterance is intended for Neon.
        Server utilizes current conversation, otherwise wake-word status
        and message "Neon" parameter used
        """
        if not self._neon_core:
            return True

        from neon_utils.message_utils import request_for_neon
        ww_enabled = self.config_core.get("listener",
                                          {}).get("wake_word_enabled", True)
        # TODO: Listen for WW state changes on the bus
        return request_for_neon(message, "neon", self.voc_match, ww_enabled)

    def report_metric(self, name, data):
        """Report a skill metric to the Mycroft servers.

        Arguments:
            name (str): Name of metric. Must use only letters and hyphens
            data (dict): JSON dictionary to report. Must be valid JSON
        """
        combined = deepcopy(data)
        combined["name"] = name
        self.bus.emit(Message("neon.metric", combined))

    def send_email(self, title, body, message=None, email_addr=None,
                   attachments=None):
        """
        Send an email to the registered user's email.
        Method here for backwards compatibility with Mycroft skills.
        Email address priority: email_addr, user prefs from message,
         fallback to DeviceApi for Mycroft method

        Arguments:
            title (str): Title of email
            body  (str): HTML body of email. This supports
                         simple HTML like bold and italics
            email_addr (str): Optional email address to send message to
            attachments (dict): Optional dict of file names to Base64 encoded files
            message (Message): Optional message to get email from
        """
        try:
            from neon_utils.mq_utils import send_mq_request
        except ImportError:
            LOG.warning("MQ Dependencies not installed")
            send_mq_request = None

        message = message or dig_for_message()
        if not email_addr and message:
            email_addr = get_user_prefs(message)["user"].get("email")

        if email_addr and send_mq_request:
            LOG.info("Send email via Neon Server")
            request_data = {"recipient": email_addr,
                            "subject": title,
                            "body": body,
                            "attachments": attachments}
            data = send_mq_request("/neon_emails", request_data,
                                   "neon_emails_input")
            return data.get("success")
        else:
            LOG.warning("Attempting to send email via Mycroft Backend")
            super().send_email(title, body)

    def make_active(self, duration_minutes=5):
        """Bump skill to active_skill list in intent_service.

        This enables converse method to be called even without skill being
        used in last 5 minutes.
        :param duration_minutes: duration in minutes for skill to remain active
         (-1 for infinite)
        """
        self.bus.emit(Message("active_skill_request",
                              {"skill_id": self.skill_id,
                               "timeout": duration_minutes}))

    def schedule_event(self, handler, when, data=None, name=None, context=None):
        # TODO: should 'when' already be a datetime? DM
        if isinstance(when, int) or isinstance(when, float):
            from datetime import datetime as dt, timedelta
            when = to_system_time(dt.now(self.sys_tz)) + timedelta(seconds=when)
            LOG.debug(f"Made a datetime: {when}")
        super().schedule_event(handler, when, data, name, context)

    @deprecated("This method is deprecated", "2.0.0")
    def request_check_timeout(self, time_wait: int,
                              intent_to_check: List[str]):
        """
        Set the specified intent to be disabled after the specified time
        :param time_wait: Time in seconds to wait before deactivating intent
        :param intent_to_check: list of intents to disable
        """
        # TODO: Consider unit tests or deprecation of this method DM
        LOG.debug(time_wait)
        LOG.debug(intent_to_check)
        if isinstance(intent_to_check, str):
            LOG.warning(f"Casting string to list: {intent_to_check}")
            intent_to_check = [intent_to_check]

        for intent in intent_to_check:
            data = {'time_out': time_wait,
                    'intent_to_check': f"{self.skill_id}:{intent}"}
            LOG.debug(f"Set Timeout: {data}")
            self.bus.emit(Message("set_timeout", data))

    def update_cached_data(self, filename: str, new_element: Any):
        """
        Updates a generic cache file
        :param filename: filename of cache object to update (relative to cacheDir)
        :param new_element: object to cache at passed location
        """
        # TODO: Move to static function with XDG compat.
        with open(os.path.join(self.cache_loc, filename), 'wb+') as file_to_update:
            pickle.dump(new_element, file_to_update, protocol=pickle.HIGHEST_PROTOCOL)

    def get_cached_data(self, filename: str,
                        file_loc: Optional[str] = None) -> dict:
        """
        Retrieves cache data from a file created/updated with update_cached_data
        :param filename: (str) filename of cache object to update
        :param file_loc: (str) path to directory containing filename (defaults to cache dir)
        :return: (dict) cache data
        """
        # TODO: Move to static function with XDG compat.
        if not file_loc:
            file_loc = self.cache_loc
        cached_location = os.path.join(file_loc, filename)
        if pathlib.Path(cached_location).exists():
            with open(cached_location, 'rb') as file:
                return pickle.load(file)
        else:
            return {}

    def decorate_api_call_use_lru(self, func):
        """
        Decorate the API-call function to use LRUcache.
        NOTE: the wrapper adds an additional argument, so decorated functions MUST be called with it!

        from wikipedia_for_humans import summary
        summary = decorate_api_call_use_lru(summary)
        result = summary(lru_query='neon', query='neon', lang='en')

        Args:
            func: the function to be decorated
        Returns: decorated function
        """
        @wraps(func)
        def wrapper(lru_query: str, *args, **kwargs):
            # TODO might use an abstract method for cached API call to define a signature
            result = self.lru_cache.get(lru_query)
            if not result:
                result = func(*args, **kwargs)
                self.lru_cache.put(key=lru_query, value=result)
            return result
        return wrapper

    def _write_cache_on_disk(self):
        """
        Write the cache on disk, reset the cache and reschedule the event.
        This handler is enabled by scheduling an event in NeonSkill.initialize().
        Returns:
        """
        filename = f"lru_{self.skill_id}"
        data_load = self.lru_cache.jsonify()
        self.update_cached_data(filename=filename, new_element=data_load)
        self.lru_cache.clear()
        self.schedule_event(self._write_cache_on_disk, CACHE_TIME_OFFSET, name="neon.load_cache_on_disk")
        return

    def _register_chat_handler(self, name: str, method: callable):
        """
        Register a chat handler entrypoint. Decorated methods must
        return a string response that will be emitted as a response to the
        incoming Message.
        :param name: name of the chat handler
        :param method: method to handle incoming chat Messages
        """

        def wrapped_handler(message):
            response = method(message)
            self.bus.emit(message.response(data={'response': response},
                                           context={'skill_id': self.skill_id}))

        self.add_event(f'chat.{name}', wrapped_handler)
        msg = dig_for_message() or Message("",
                                           context={'skill_id': self.skill_id})
        self.bus.emit(msg.forward("register_chat_handler", {'name': name}))

    def _register_decorated(self):
        for attr_name in get_non_properties(self):
            method = getattr(self, attr_name)
            if hasattr(method, 'intents'):
                for intent in getattr(method, 'intents'):
                    self.register_intent(intent, method)

            if hasattr(method, 'intent_files'):
                for intent_file in getattr(method, 'intent_files'):
                    self.register_intent_file(intent_file, method)

            if hasattr(method, 'chat_handler'):
                self._register_chat_handler(getattr(method, 'chat_handler'),
                                            method)

    @property
    def location(self):
        """
        Backwards-compatible location property. Returns core location config if
        user location isn't specified.
        """
        from neon_utils.configuration_utils import get_mycroft_compatible_location
        return get_mycroft_compatible_location(get_user_prefs()["location"])

    def _init_settings(self):
        """
        Extends the default method to handle settingsmeta defaults locally
        """
        from neon_utils.configuration_utils import dict_update_keys
        OVOSSkill._init_settings(self)
        settings_from_disk = dict(self.settings)
        dict_update_keys(self._settings, self._read_default_settings())
        if self._settings != settings_from_disk:
            LOG.info("Updated default settings from skill metadata")
            self._settings.store()
            self._initial_settings = dict(self._settings)
        LOG.info(f"Skill initialized with settings: {self.settings}")

    def _handle_converse_request(self, message: Message):
        # TODO: Remove patch after ovos-core 0.0.8
        if message.msg_type == "skill.converse.request" and \
                message.data.get('skill_id') != self.skill_id:
            # Legacy request not for Neon
            return
        if message.msg_type == "skill.converse.request":
            message.msg_type = "neon.converse.request"
        OVOSSkill._handle_converse_request(self, message)

    def _read_default_settings(self):
        from neon_utils.configuration_utils import parse_skill_default_settings
        yaml_path = os.path.join(self.root_dir, "settingsmeta.yml")
        json_path = os.path.join(self.root_dir, "settingsmeta.json")
        if os.path.isfile(yaml_path):
            with open(yaml_path) as f:
                self.settings_meta = yaml.safe_load(f) or dict()
        elif os.path.isfile(json_path):
            with open(json_path) as f:
                self.settings_meta = json.load(f)
        else:
            return dict()
        return parse_skill_default_settings(self.settings_meta)

    @resolve_message
    def speak(self, utterance, expect_response=False, wait=False, meta=None,
              message=None, private=False, speaker=None):
        """
        Speak an utterance.
        Arguments:
            utterance (str):        sentence mycroft should speak
            expect_response (bool): set to True if Mycroft should listen for a response immediately after
                                    speaking the utterance.
            wait (bool):            set to True to block while the text is being spoken.
            meta:                   Information of what built the sentence.
            message (Message):      message associated with the input that this speak is associated with
            private (bool):         flag to indicate this message contains data that is private to the requesting user
            speaker (dict):         dict containing language or voice data to override user preference values

        """
        from neon_utils.signal_utils import check_for_signal, wait_for_signal_clear
        # registers the skill as being active
        meta = meta or {}
        meta['skill'] = self.name
        self.enclosure.register(self.name)
        if utterance:
            if not message:
                LOG.debug('message is None.')
                message = Message("speak")
            if not speaker:
                speaker = message.data.get("speaker", None)

            nick = get_message_user(message)

            if private and message.context.get("klat_data"):
                LOG.debug("Private Message")
                title = message.context["klat_data"].get("title") or \
                    "!PRIVATE:Neon"
                need_at_sign = True
                if title.startswith("!PRIVATE"):
                    users = title.split(':')[1].split(',')
                    for idx, val in enumerate(users):
                        users[idx] = val.strip()
                    if len(users) == 2 and "Neon" in users:
                        need_at_sign = False
                    elif len(users) == 1:
                        need_at_sign = False
                    elif nick.startswith("guest"):
                        need_at_sign = False
                if need_at_sign:
                    LOG.debug("Send message to private cid!")
                    utterance = f"@{nick} {utterance}"

            data = {"utterance": utterance,
                    "lang": self.lang,
                    "expect_response": expect_response,
                    "meta": meta,
                    "speaker": speaker,
                    "speak_ident": str(time.time())}

            if message.context.get("cc_data", {}).get("emit_response"):
                msg_to_emit = message.reply("skills:execute.response", data,
                                            {"destination": ["skills"],
                                             "source": ["skills"]})
            else:
                message.context.get("timing", {})["speech_start"] = time.time()
                msg_to_emit = message.reply("speak", data,
                                            {"destination": ["skills"],
                                             "source": ["audio"]})
                LOG.debug(f"Skill speak! {data}")
            LOG.debug(msg_to_emit.msg_type)

            if wait and check_for_signal("neon_speak_api", -1):
                self.bus.wait_for_response(msg_to_emit,
                                           msg_to_emit.data['speak_ident'],
                                           self._speak_timeout)
            else:
                self.bus.emit(msg_to_emit)
                if wait and not message.context.get("klat_data"):
                    LOG.debug("Using legacy isSpeaking signal")
                    wait_for_signal_clear('isSpeaking')

        else:
            LOG.warning("Null utterance passed to speak")
            LOG.warning(f"{self.name} | message={message}")

    @resolve_message
    def speak_dialog(self, key, data=None, expect_response=False, wait=False,
                     message=None, private=False, speaker=None):
        """ Speak a random sentence from a dialog file.

        Arguments:
            :param key: dialog file key (e.g. "hello" to speak from the file
                "locale/en-us/hello.dialog")
            :param data: information used to populate key
            :param expect_response: set to True if Mycroft should listen for a
                response immediately after speaking.
            :param wait: set to True to block while the text is being spoken.
            :param message: associated message from request
            :param private: private flag (server use only)
            :param speaker: optional dict of speaker info to use
        """
        data = data or {}
        LOG.debug(f"data={data}")
        if self.dialog_renderer:  # TODO: Pass index (0) here to use non-random responses DM
            to_speak = self.dialog_renderer.render(key, data)
        else:
            to_speak = key
        self.speak(to_speak,
                   expect_response, message=message, private=private,
                   speaker=speaker, wait=wait, meta={'dialog': key,
                                                     'data': data})

    @resolve_message
    def get_response(self, dialog: str = '', data: Optional[dict] = None,
                     validator=None, on_fail=None, num_retries: int = -1,
                     message: Optional[Message] = None) -> Optional[str]:
        """
        Gets a response from a user. Speaks the passed dialog file or string
        and then optionally plays a listening confirmation sound and
        starts listening if in wake words mode.
        Wraps the default Mycroft method to add support for multiple users and
        running without a wake word.

        Arguments:
            dialog (str): Optional dialog to speak to the user
            data (dict): Data used to render the dialog
            validator (any): Function with following signature
                def validator(utterance):
                    return utterance != "red"
            on_fail (any): Dialog or function returning literal string
                           to speak on invalid input.  For example:
                def on_fail(utterance):
                    return "nobody likes the color red, pick another"
            num_retries (int): Times to ask user for input, -1 for infinite
                NOTE: User can not respond and timeout or say "cancel" to stop
            message (Message): Message associated with request

        Returns:
            str: User's reply or None if timed out or canceled
        """
        user = get_message_user(message) or "local" if message else "local"
        data = data or {}

        def on_fail_default(utterance):
            fail_data = data.copy()
            fail_data['utterance'] = utterance

            if on_fail:
                to_speak = on_fail
            else:
                to_speak = dialog
            if self.dialog_renderer:
                return self.dialog_renderer.render(to_speak, data)
            else:
                return to_speak

        def is_cancel(utterance):
            return self.voc_match(utterance, 'cancel')

        def validator_default(utterance):
            # accept anything except 'cancel'
            return not is_cancel(utterance)

        on_fail_fn = on_fail if callable(on_fail) else on_fail_default
        validator = validator or validator_default

        # Ensure we have a message to forward
        message = message or dig_for_message()
        if not message:
            LOG.warning(f"Could not locate message associated with request!")
            message = Message("get_response")

        # If skill has dialog, render the input
        if self.dialog_renderer:
            dialog = self.dialog_renderer.render(dialog, data)

        if dialog:
            self.speak(dialog, wait=True, message=message, private=True)
        self.bus.emit(message.forward('mycroft.mic.listen'))
        return self._wait_response(is_cancel, validator, on_fail_fn,
                                   num_retries, message, user)

    def _wait_response(self, is_cancel, validator, on_fail, num_retries,
                       message=None, user: str = None):
        """
        Loop until a valid response is received from the user or the retry
        limit is reached.

        Arguments:
            is_cancel (callable): function checking cancel criteria
            validator (callable): function checking for a valid response
            on_fail (callable): function handling retries
            message (Message): message associated with request
        """
        user = user or "local"
        num_fails = 0
        while True:
            response = self.__get_response(user)

            if response is None:  # No Response
                # if nothing said, only prompt one more time
                num_none_fails = 1 if num_retries < 0 else num_retries
                LOG.debug(f"num_none_fails={num_none_fails}|"
                          f"num_fails={num_fails}")
                if num_fails >= num_none_fails:
                    LOG.info("No user response")
                    return None
            else:  # Some response
                # catch user saying 'cancel'
                if is_cancel(response):
                    LOG.info("User cancelled")
                    return None
                validated = validator(response)
                # returns the validated value or the response
                # (backwards compat)
                if validated is not False and validated is not None:
                    LOG.debug(f"Returning validated response")
                    return response if validated is True else validated
                LOG.debug(f"User response not validated: {response}")
            # Unvalidated or no response
            num_fails += 1
            if 0 < num_retries < num_fails:
                LOG.info(f"Failed ({num_fails}) through all retries "
                         f"({num_retries})")
                return None

            # Validation failed, retry
            line = on_fail(response)
            if line:
                LOG.debug(f"Speaking failure dialog: {line}")
                self.speak(line, wait=True, message=message, private=True)

            LOG.debug("Listen for another response")
            msg = message.reply('mycroft.mic.listen') or \
                Message('mycroft.mic.listen',
                        context={"skill_id": self.skill_id})
            self.bus.emit(msg)

    def __get_response(self, user="local"):
        """
        Helper to get a response from the user

        Arguments:
            user (str): user associated with response
        Returns:
            str: user's response or None on a timeout
        """
        event = Event()

        def converse(message):
            resp_user = get_message_user(message) or "local"
            if resp_user == user:
                utterances = message.data.get("utterances")
                converse.response = utterances[0] if utterances else None
                event.set()
                LOG.info(f"Got response: {converse.response}")
                return True
            LOG.debug(f"Ignoring input from: {resp_user}")
            return False

        # install a temporary conversation handler
        self.make_active()
        converse.response = None
        default_converse = self.converse
        self.converse = converse

        if not event.wait(self._get_response_timeout):
            LOG.warning("Timed out waiting for user response")
        self.converse = default_converse
        return converse.response

    # renamed in base class for naming consistency
    # refactored to use new resource utils
    def translate(self, text: str, data: Optional[dict] = None):
        """
        Deprecated method for translating a dialog file.
        use self.resources.render_dialog(text, data) instead
        """
        log_deprecation("Use `resources.render_dialog`", "2.0.0")
        return self.resources.render_dialog(text, data)

    # renamed in base class for naming consistency
    # refactored to use new resource utils
    def translate_namedvalues(self, name: str, delim: str = ','):
        """
        Deprecated method for translating a name/value file.
        use self.resources.load_named_value_filetext, data) instead
        """
        log_deprecation("Use `resources.load_named_value_file`", "2.0.0")
        return self.resources.load_named_value_file(name, delim)

    # renamed in base class for naming consistency
    # refactored to use new resource utils
    def translate_list(self, list_name: str, data: Optional[dict] = None):
        """
        Deprecated method for translating a list.
        use delf.resources.load_list_file(text, data) instead
        """
        log_deprecation("Use `resources.load_list_file`", "2.0.0")
        return self.resources.load_list_file(list_name, data)

    # renamed in base class for naming consistency
    # refactored to use new resource utils
    def translate_template(self, template_name: str,
                           data: Optional[dict] = None):
        """
        Deprecated method for translating a template file
        use self.resources.template_file(text, data) instead
        """
        log_deprecation("Use `resources.template_file`", "2.0.0")
        return self.resources.load_template_file(template_name, data)

    def init_dialog(self, root_directory: Optional[str] = None):
        """
        DEPRECATED: use load_dialog_files instead
        """
        log_deprecation("Use `load_dialog_files`", "2.0.0")
        self.load_dialog_files(root_directory)

    def add_event(self, name: str, handler: callable,
                  handler_info: Optional[str] = None, once: bool = False,
                  speak_errors: bool = True):
        # TODO: Remove with ovos-workshop==0.0.13
        try:
            # Patching FakeBus compat. with MessageBusClient
            if hasattr(self.bus, "ee"):
                emitter = self.bus.ee
            else:
                emitter = self.bus.emitter
            if handler_info == "mycroft.skill.handler" and \
                    emitter.listeners(name):
                LOG.warning(f"Not re-registering intent handler {name}")
                return
        except Exception as e:
            LOG.exception(e)
        OVOSSkill.add_event(self, name, handler, handler_info, once,
                            speak_errors)




# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import requests
import json

from os import makedirs
from os.path import join, isfile, isdir, dirname
from time import time
from ovos_utils.log import LOG
from ovos_utils.xdg_utils import xdg_cache_home

_DEFAULT_BACKEND_URL = "https://hana.neonaiservices.com"
_client_config = {}
_headers = {}


def _get_client_config_path(url: str = _DEFAULT_BACKEND_URL):
    url_key = hash(url)
    return join(xdg_cache_home(), "neon", f"hana_token_{url_key}.json")


class ServerException(Exception):
    """Exception class representing a backend server communication error"""


def _init_client(backend_address: str):
    """
    Initialize request headers for making backend requests. If a local cache is
    available it will be used, otherwise an auth request will be made to the
    specified backend server
    @param backend_address: Hana server URL to connect to
    """
    global _client_config
    global _headers

    if not _client_config:
        client_config_path = _get_client_config_path(backend_address)
        if isfile(client_config_path):
            with open(client_config_path) as f:
                _client_config = json.load(f)
        else:
            _get_token(backend_address)

    if not _headers:
        _headers = {"Authorization": f"Bearer {_client_config['access_token']}"}


def _get_token(backend_address: str, username: str = "guest",
               password: str = "password"):
    """
    Get new auth tokens from the specified server. This will cache the returned
    token, overwriting any previous data at the cache path.
    @param backend_address: Hana server URL to connect to
    @param username: Username to authorize
    @param password: Password for specified username
    """
    global _client_config
    # TODO: username/password from configuration
    resp = requests.post(f"{backend_address}/auth/login",
                         json={"username": username,
                               "password": password})
    if not resp.ok:
        raise ServerException(f"Error logging into {backend_address}. "
                              f"{resp.status_code}: {resp.text}")
    _client_config = resp.json()
    client_config_path = _get_client_config_path(backend_address)
    if not isdir(dirname(client_config_path)):
        makedirs(dirname(client_config_path))
    with open(client_config_path, "w+") as f:
        json.dump(_client_config, f, indent=2)


def _refresh_token(backend_address: str):
    """
    Get new tokens from the specified server using an existing refresh token
    (if it exists). This will update the cached tokens and associated metadata.
    @param backend_address: Hana server URL to connect to
    """
    global _client_config
    _init_client(backend_address)
    update = requests.post(f"{backend_address}/auth/refresh", json={
        "access_token": _client_config.get("access_token"),
        "refresh_token": _client_config.get("refresh_token"),
        "client_id": _client_config.get("client_id")})
    if not update.ok:
        raise ServerException(f"Error updating token from {backend_address}. "
                              f"{update.status_code}: {update.text}")
    _client_config = update.json()
    client_config_path = _get_client_config_path(backend_address)
    with open(client_config_path, "w+") as f:
        json.dump(_client_config, f, indent=2)


def request_backend(endpoint: str, request_data: dict,
                    server_url: str = _DEFAULT_BACKEND_URL) -> dict:
    """
    Make a request to a Hana backend server and return the json response
    @param endpoint: server endpoint to query
    @param request_data: dict data to send in request body
    @param server_url: Base URL of Hana server to query
    @returns: dict response
    """
    _init_client(server_url)
    if time() >= _client_config.get("expiration", 0):
        try:
            _refresh_token(server_url)
        except ServerException as e:
            LOG.error(e)
            _get_token(server_url)
    resp = requests.post(f"{server_url}/{endpoint.lstrip('/')}",
                         json=request_data, headers=_headers)
    if resp.ok:
        return resp.json()
    else:
        raise ServerException(f"Error response {resp.status_code}: {resp.text}")



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import unittest
import sys
import os

from ovos_utils.messagebus import FakeBus

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
from neon_utils.validator_utils import *


class ValidatorUtilTests(unittest.TestCase):
    def test_numeric_confirmation_validator_valid(self):
        validator = numeric_confirmation_validator("123")
        self.assertTrue(validator("123"))
        self.assertTrue(validator("1 2 3."))
        self.assertTrue(validator("one two three"))
        self.assertTrue(validator("one hundred twenty three"))
        self.assertTrue(validator("blah blah one two three nah"))

        self.assertFalse(validator("one thousand two hundred and three"))
        self.assertFalse(validator("123 4"))

    def test_numeric_confirmation_validator_type_error(self):
        with self.assertRaises(TypeError):
            numeric_confirmation_validator(123)

        with self.assertRaises(ValueError):
            numeric_confirmation_validator("one two three")

        with self.assertRaises(ValueError):
            numeric_confirmation_validator("")

    def test_string_confirmation_validator_valid(self):
        validator = string_confirmation_validator("test phrase")
        self.assertTrue(validator("test phrase"))
        self.assertTrue(validator("yes test phrase yes"))

        self.assertFalse(validator("test no phrase"))
        self.assertFalse(validator("phrase test"))

    def test_string_confirmation_validator_type_error(self):
        with self.assertRaises(TypeError):
            string_confirmation_validator(123)

        with self.assertRaises(ValueError):
            string_confirmation_validator("")

    def test_voc_confirmation_validator_valid(self):
        sys.path.append(os.path.dirname(__file__))
        from valid_skill import ValidNeonSkill
        skill = ValidNeonSkill()
        if hasattr(skill, "_startup"):
            skill._startup(FakeBus())
        validator = voc_confirmation_validator("test", ValidNeonSkill())
        self.assertTrue(validator("test"))
        self.assertTrue(validator("something"))

        self.assertFalse(validator("else"))
        self.assertFalse(validator("false"))

    def test_voc_confirmation_validator_type_error(self):
        sys.path.append(os.path.dirname(__file__))
        from valid_skill import ValidNeonSkill
        skill = ValidNeonSkill()
        with self.assertRaises(TypeError):
            voc_confirmation_validator(123, skill)

        with self.assertRaises(FileNotFoundError):
            voc_confirmation_validator("invalid", skill)


if __name__ == '__main__':
    unittest.main()



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Copyright 2018 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from enum import Enum, IntEnum
from abc import ABC, abstractmethod
from ovos_bus_client import Message
from ovos_utils.log import log_deprecation

from neon_utils.skills.neon_skill import NeonSkill

from ovos_utils.skills.audioservice import AudioServiceInterface as AudioService


class CPSMatchLevel(Enum):
    EXACT = 1
    MULTI_KEY = 2
    TITLE = 3
    ARTIST = 4
    CATEGORY = 5
    GENERIC = 6


class CPSTrackStatus(IntEnum):
    DISAMBIGUATION = 1  # not queued for playback, show in gui
    PLAYING = 20  # Skill is handling playback internally
    PLAYING_AUDIOSERVICE = 21  # Skill forwarded playback to audio service
    PLAYING_GUI = 22  # Skill forwarded playback to gui
    PLAYING_ENCLOSURE = 23  # Skill forwarded playback to enclosure
    QUEUED = 30  # Waiting playback to be handled inside skill
    QUEUED_AUDIOSERVICE = 31  # Waiting playback in audio service
    QUEUED_GUI = 32  # Waiting playback in gui
    QUEUED_ENCLOSURE = 33  # Waiting for playback in enclosure
    PAUSED = 40  # media paused but ready to resume
    STALLED = 60  # playback has stalled, reason may be unknown
    BUFFERING = 61  # media is buffering from an external source
    END_OF_MEDIA = 90  # playback finished, is the default state when CPS loads


class CommonPlaySkill(NeonSkill, ABC):
    """ To integrate with the common play infrastructure of Mycroft
    skills should use this base class and override the two methods
    `CPS_match_query_phrase` (for checking if the skill can play the
    utterance) and `CPS_start` for launching the media.

    The class makes the skill available to queries from the
    mycroft-playback-control skill and no special vocab for starting playback
    is needed.
    """
    def __init__(self, *args, **kwargs):
        log_deprecation("This base class is deprecated. Implement "
                        "`ovos_workshop.skills.common_play."
                        "OVOSCommonPlaybackSkill`", "2.0.0")
        NeonSkill.__init__(self, *args, **kwargs)
        self.audioservice = None
        self.play_service_string = None

        # "MusicServiceSkill" -> "Music Service"
        spoken = kwargs.get("name") or args[0] or self.__class__.__name__
        self.spoken_name = re.sub(r"([a-z])([A-Z])", r"\g<1> \g<2>",
                                  spoken.replace("Skill", ""))
        # NOTE: Derived skills will likely want to override self.spoken_name
        # with a translatable name in their initialize() method.

    def bind(self, bus):
        """Overrides the normal bind method.

        Adds handlers for play:query and play:start messages allowing
        interaction with the playback control skill.

        This is called automatically during setup, and
        need not otherwise be used.
        """
        if bus:
            super().bind(bus)
            self.audioservice = AudioService(self.bus)
            self.add_event('play:query', self.__handle_play_query)
            self.add_event('play:start', self.__handle_play_start)

    def __handle_play_query(self, message):
        """Query skill if it can start playback from given phrase."""
        search_phrase = message.data["phrase"]

        # First, notify the requestor that we are attempting to handle
        # (this extends a timeout while this skill looks for a match)
        self.bus.emit(message.response({"phrase": search_phrase,
                                        "skill_id": self.skill_id,
                                        "searching": True}))

        # Now invoke the CPS handler to let the skill perform its search
        result = self.CPS_match_query_phrase(search_phrase, message)

        if result:
            match = result[0]
            level = result[1]
            callback = result[2] if len(result) > 2 else None
            confidence = self.__calc_confidence(match, search_phrase, level)
            self.bus.emit(message.response({"phrase": search_phrase,
                                            "skill_id": self.skill_id,
                                            "callback_data": callback,
                                            "service_name": self.spoken_name,
                                            "conf": confidence}))
        else:
            # Signal we are done (can't handle it)
            self.bus.emit(message.response({"phrase": search_phrase,
                                            "skill_id": self.skill_id,
                                            "searching": False}))

    def __calc_confidence(self, match, phrase, level):
        """Translate confidence level and match to a 0-1 value.

        "play pandora"
        "play pandora is my girlfriend"
        "play tom waits on pandora"

        Assume the more of the words that get consumed, the better the match

        Arguments:
            match (str): Matching string
            phrase (str): original input phrase
            level (CPSMatchLevel): match level
        """
        consumed_pct = len(match.split()) / len(phrase.split())
        if consumed_pct > 1.0:
            consumed_pct = 1.0 / consumed_pct  # deal with over/under-matching

        # We'll use this to modify the level, but don't want it to allow a
        # match to jump to the next match level.  So bonus is 0 - 0.05 (1/20)
        bonus = consumed_pct / 20.0

        if level == CPSMatchLevel.EXACT:
            return 1.0
        elif level == CPSMatchLevel.MULTI_KEY:
            return 0.9 + bonus
        elif level == CPSMatchLevel.TITLE:
            return 0.8 + bonus
        elif level == CPSMatchLevel.ARTIST:
            return 0.7 + bonus
        elif level == CPSMatchLevel.CATEGORY:
            return 0.6 + bonus
        elif level == CPSMatchLevel.GENERIC:
            return 0.5 + bonus
        else:
            return 0.0  # should never happen

    def __handle_play_start(self, message):
        """Bus handler for starting playback using the skill."""
        if message.data["skill_id"] != self.skill_id:
            # Not for this skill!
            return
        phrase = message.data["phrase"]
        data = message.data.get("callback_data")

        # Stop any currently playing audio
        if self.audioservice.is_playing:
            self.audioservice.stop()
        self.bus.emit(message.forward("mycroft.stop"))
        # TODO: maybe do this with some param to not clear gui in CP skills DM

        # Save for CPS_play() later, e.g. if phrase includes modifiers like
        # "... on the chromecast"
        self.play_service_string = phrase

        # Invoke derived class to provide playback data
        self.CPS_start(phrase, data, message)

    def CPS_play(self, *args, **kwargs):
        """Begin playback of a media file or stream

        Normally this method will be invoked with somthing like:
           self.CPS_play(url)
        Advanced use can also include keyword arguments, such as:
           self.CPS_play(url, repeat=True)

        Args:
            same as the Audioservice.play method
        """
        # Inject the user's utterance in case the audio backend wants to
        # interpret it.  E.g. "play some rock at full volume on the stereo"
        if 'utterance' not in kwargs:
            kwargs['utterance'] = self.play_service_string
        # TODO: Enable WW, set sink name, set signal to handle WW on stop (like AVMusic) DM
        #       Move gui display to here from Playback-Control skill? DM
        self.audioservice.play(*args, **kwargs)
        self.CPS_send_status(uri=args[0],
                             status=CPSTrackStatus.PLAYING_AUDIOSERVICE)

    def stop(self):
        """Stop anything playing on the audioservice."""
        if self.audioservice.is_playing:
            self.audioservice.stop()
            return True
        else:
            return False

    ######################################################################
    # Abstract methods
    # All of the following must be implemented by a skill that wants to
    # act as a CommonPlay Skill
    @abstractmethod
    def CPS_match_query_phrase(self, phrase, message):
        """Analyze phrase to see if it is a play-able phrase with this skill.

        Arguments:
            phrase (str): User phrase uttered after "Play", e.g. "some music"
            message (Message): Message associated with request
        Returns:
            (match, CPSMatchLevel[, callback_data]) or None: Tuple containing
                 a string with the appropriate matching phrase, the PlayMatch
                 type, and optionally data to return in the callback if the
                 match is selected.
        """
        # Derived classes must implement this, e.g.
        #
        # if phrase in ["Zoosh"]:
        #     return ("Zoosh", CPSMatchLevel.Generic, {"hint": "music"})
        # or:
        # zoosh_song = find_zoosh(phrase)
        # if zoosh_song and "Zoosh" in phrase:
        #     # "play Happy Birthday in Zoosh"
        #     return ("Zoosh", CPSMatchLevel.MULTI_KEY, {"song": zoosh_song})
        # elif zoosh_song:
        #     # "play Happy Birthday"
        #     return ("Zoosh", CPSMatchLevel.TITLE, {"song": zoosh_song})
        # elif "Zoosh" in phrase
        #     # "play Zoosh"
        #     return ("Zoosh", CPSMatchLevel.GENERIC, {"cmd": "random"})
        return None

    @abstractmethod
    def CPS_start(self, phrase, data, message=None):
        """Begin playing whatever is specified in 'phrase'

        Arguments:
            phrase (str): User phrase uttered after "Play", e.g. "some music"
            data (dict): Callback data specified in match_query_phrase()
        """
        # Derived classes must implement this, e.g.
        # self.CPS_play("http://zoosh.com/stream_music")
        pass

    def CPS_extend_timeout(self, timeout=5):
        """Request Common Play Framework to wait another {timeout} seconds
        for an answer from this skill.

        Arguments:
            timeout (int): Number of seconds
        """
        self.bus.emit(Message('play:query.response',
                              {"phrase": self.play_service_string,
                               "searching": True,
                               "timeout": timeout,
                               "skill_id": self.skill_id}))

    def CPS_send_status(self, artist='', track='', album='', image='',
                        uri='', track_length=None, elapsed_time=None,
                        playlist_position=None,
                        status=CPSTrackStatus.DISAMBIGUATION, **kwargs):
        """Inform system of playback status.

        If a skill is handling playback and wants the playback control to be
        aware of it's current status it can emit this message indicating that
        it's performing playback and can provide some standard info.

        All parameters are optional so any can be left out. Also if extra
        non-standard parameters are added, they too will be sent in the message
        data.

        Arguments:
            artist (str): Current track artist
            track (str): Track name
            album (str): Album title
            image (str): url for image to show
            uri (str): uri for track
            track_length (float): track length in seconds
            elapsed_time (float): current offset into track in seconds
            playlist_position (int): Position in playlist of current track
        """
        data = {'skill': self.name,
                'uri': uri,
                'artist': artist,
                'album': album,
                'track': track,
                'image': image,
                'track_length': track_length,
                'elapsed_time': elapsed_time,
                'playlist_position': playlist_position,
                'status': status
                }
        data = {**data, **kwargs}  # Merge extra arguments
        self.bus.emit(Message('play:status', data))

    def CPS_send_tracklist(self, tracklist):
        """Inform system of playlist track info.

        Provides track data for playlist

        Arguments:
            tracklist (list/dict): Tracklist data
        """
        tracklist = tracklist or []
        if not isinstance(tracklist, list):
            tracklist = [tracklist]
        for idx, track in enumerate(tracklist):
            self.CPS_send_status(playlist_position=idx, **track)



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json
import time
import os.path
import yaml

from threading import Event
from typing import Optional
from json_database import JsonStorage
from ovos_bus_client.message import Message
from ovos_utils.log import log_deprecation
from ovos_workshop.skills.mycroft_skill import MycroftSkill
from ovos_utils.skills.settings import get_local_settings

from neon_utils.signal_utils import wait_for_signal_clear, check_for_signal
from neon_utils.logger import LOG
from neon_utils.message_utils import get_message_user, dig_for_message, resolve_message
from neon_utils.configuration_utils import dict_update_keys, \
    parse_skill_default_settings, get_mycroft_compatible_location
from neon_utils.user_utils import get_user_prefs


class PatchedMycroftSkill(MycroftSkill):
    def __init__(self, name=None, bus=None, *args, **kwargs):
        MycroftSkill.__init__(self, name, bus, *args, **kwargs)
        log_deprecation("This base class is deprecated. Implement either"
                        "`NeonSkill` or `OVOSSkill`", "2.0.0")
        # TODO: Should below defaults be global config?
        # allow skills to specify timeout overrides per-skill
        self._speak_timeout = 30
        self._get_response_timeout = 15  # 10 for listener, 5 for STT, then timeout

    @property
    def location(self):
        """
        Backwards-compatible location property. Returns core location config if
        user location isn't specified.
        """
        return get_mycroft_compatible_location(get_user_prefs()["location"])

    def _init_settings(self):
        """
        Extends the default method to handle settingsmeta defaults locally
        """
        super()._init_settings()
        skill_settings = get_local_settings(self.settings_path)
        settings_from_disk = dict(skill_settings)
        self.settings = dict_update_keys(skill_settings,
                                         self._read_default_settings())
        if self.settings != settings_from_disk:
            if isinstance(self.settings, JsonStorage):
                self.settings.store()
            else:
                with open(self.settings_path, "w+") as f:
                    json.dump(self.settings, f, indent=4)
        self._initial_settings = dict(self.settings)

    def _read_default_settings(self):
        yaml_path = os.path.join(self.root_dir, "settingsmeta.yml")
        json_path = os.path.join(self.root_dir, "settingsmeta.json")
        if os.path.isfile(yaml_path):
            with open(yaml_path) as f:
                self.settings_meta = yaml.safe_load(f) or dict()
        elif os.path.isfile(json_path):
            with open(json_path) as f:
                self.settings_meta = json.load(f)
        else:
            return dict()
        return parse_skill_default_settings(self.settings_meta)

    @resolve_message
    def speak(self, utterance, expect_response=False, wait=False, meta=None,
              message=None, private=False, speaker=None):
        """
        Speak an utterance.
        Arguments:
            utterance (str):        sentence mycroft should speak
            expect_response (bool): set to True if Mycroft should listen for a response immediately after
                                    speaking the utterance.
            wait (bool):            set to True to block while the text is being spoken.
            meta:                   Information of what built the sentence.
            message (Message):      message associated with the input that this speak is associated with
            private (bool):         flag to indicate this message contains data that is private to the requesting user
            speaker (dict):         dict containing language or voice data to override user preference values

        """
        # registers the skill as being active
        meta = meta or {}
        meta['skill'] = self.name
        self.enclosure.register(self.name)
        if utterance:
            if not message:
                LOG.debug('message is None.')
                message = Message("speak")
            if not speaker:
                speaker = message.data.get("speaker", None)

            nick = get_message_user(message)

            if private and message.context.get("klat_data"):
                LOG.debug("Private Message")
                title = message.context["klat_data"].get("title") or \
                    "!PRIVATE:Neon"
                need_at_sign = True
                if title.startswith("!PRIVATE"):
                    users = title.split(':')[1].split(',')
                    for idx, val in enumerate(users):
                        users[idx] = val.strip()
                    if len(users) == 2 and "Neon" in users:
                        need_at_sign = False
                    elif len(users) == 1:
                        need_at_sign = False
                    elif nick.startswith("guest"):
                        need_at_sign = False
                if need_at_sign:
                    LOG.debug("Send message to private cid!")
                    utterance = f"@{nick} {utterance}"

            data = {"utterance": utterance,
                    "lang": self.lang,
                    "expect_response": expect_response,
                    "meta": meta,
                    "speaker": speaker,
                    "speak_ident": str(time.time())}

            if message.context.get("cc_data", {}).get("emit_response"):
                msg_to_emit = message.reply("skills:execute.response", data,
                                            {"destination": ["skills"],
                                             "source": ["skills"]})
            else:
                message.context.get("timing", {})["speech_start"] = time.time()
                msg_to_emit = message.reply("speak", data,
                                            {"destination": ["skills"],
                                             "source": ["audio"]})
                LOG.debug(f"Skill speak! {data}")
            LOG.debug(msg_to_emit.msg_type)

            if wait and check_for_signal("neon_speak_api", -1):
                self.bus.wait_for_response(msg_to_emit,
                                           msg_to_emit.data['speak_ident'],
                                           self._speak_timeout)
            else:
                self.bus.emit(msg_to_emit)
                if wait and not message.context.get("klat_data"):
                    LOG.debug("Using legacy isSpeaking signal")
                    wait_for_signal_clear('isSpeaking')

        else:
            LOG.warning("Null utterance passed to speak")
            LOG.warning(f"{self.name} | message={message}")

    @resolve_message
    def speak_dialog(self, key, data=None, expect_response=False, wait=False,
                     message=None, private=False, speaker=None):
        """ Speak a random sentence from a dialog file.

        Arguments:
            :param key: dialog file key (e.g. "hello" to speak from the file
                "locale/en-us/hello.dialog")
            :param data: information used to populate key
            :param expect_response: set to True if Mycroft should listen for a
                response immediately after speaking.
            :param wait: set to True to block while the text is being spoken.
            :param message: associated message from request
            :param private: private flag (server use only)
            :param speaker: optional dict of speaker info to use
        """
        data = data or {}
        LOG.debug(f"data={data}")
        if self.dialog_renderer:  # TODO: Pass index (0) here to use non-random responses DM
            to_speak = self.dialog_renderer.render(key, data)
        else:
            to_speak = key
        self.speak(to_speak,
                   expect_response, message=message, private=private,
                   speaker=speaker, wait=wait, meta={'dialog': key,
                                                     'data': data})

    @resolve_message
    def get_response(self, dialog: str = '', data: Optional[dict] = None,
                     validator=None, on_fail=None, num_retries: int = -1,
                     message: Optional[Message] = None) -> Optional[str]:
        """
        Gets a response from a user. Speaks the passed dialog file or string
        and then optionally plays a listening confirmation sound and
        starts listening if in wake words mode.
        Wraps the default Mycroft method to add support for multiple users and
        running without a wake word.

        Arguments:
            dialog (str): Optional dialog to speak to the user
            data (dict): Data used to render the dialog
            validator (any): Function with following signature
                def validator(utterance):
                    return utterance != "red"
            on_fail (any): Dialog or function returning literal string
                           to speak on invalid input.  For example:
                def on_fail(utterance):
                    return "nobody likes the color red, pick another"
            num_retries (int): Times to ask user for input, -1 for infinite
                NOTE: User can not respond and timeout or say "cancel" to stop
            message (Message): Message associated with request

        Returns:
            str: User's reply or None if timed out or canceled
        """
        user = get_message_user(message) or "local" if message else "local"
        data = data or {}

        def on_fail_default(utterance):
            fail_data = data.copy()
            fail_data['utterance'] = utterance

            if on_fail:
                to_speak = on_fail
            else:
                to_speak = dialog
            if self.dialog_renderer:
                return self.dialog_renderer.render(to_speak, data)
            else:
                return to_speak

        def is_cancel(utterance):
            return self.voc_match(utterance, 'cancel')

        def validator_default(utterance):
            # accept anything except 'cancel'
            return not is_cancel(utterance)

        on_fail_fn = on_fail if callable(on_fail) else on_fail_default
        validator = validator or validator_default

        # Ensure we have a message to forward
        message = message or dig_for_message()
        if not message:
            LOG.warning(f"Could not locate message associated with request!")
            message = Message("get_response")

        # If skill has dialog, render the input
        if self.dialog_renderer:
            dialog = self.dialog_renderer.render(dialog, data)

        if dialog:
            self.speak(dialog, wait=True, message=message, private=True)
        self.bus.emit(message.forward('mycroft.mic.listen'))
        return self._wait_response(is_cancel, validator, on_fail_fn,
                                   num_retries, message, user)

    def _wait_response(self, is_cancel, validator, on_fail, num_retries,
                       message=None, user: str = None):
        """
        Loop until a valid response is received from the user or the retry
        limit is reached.

        Arguments:
            is_cancel (callable): function checking cancel criteria
            validator (callable): function checking for a valid response
            on_fail (callable): function handling retries
            message (Message): message associated with request
        """
        user = user or "local"
        num_fails = 0
        while True:
            response = self.__get_response(user)

            if response is None:  # No Response
                # if nothing said, only prompt one more time
                num_none_fails = 1 if num_retries < 0 else num_retries
                LOG.debug(f"num_none_fails={num_none_fails}|"
                          f"num_fails={num_fails}")
                if num_fails >= num_none_fails:
                    LOG.info("No user response")
                    return None
            else:  # Some response
                # catch user saying 'cancel'
                if is_cancel(response):
                    LOG.info("User cancelled")
                    return None
                validated = validator(response)
                # returns the validated value or the response
                # (backwards compat)
                if validated is not False and validated is not None:
                    LOG.debug(f"Returning validated response")
                    return response if validated is True else validated
                LOG.debug(f"User response not validated: {response}")
            # Unvalidated or no response
            num_fails += 1
            if 0 < num_retries < num_fails:
                LOG.info(f"Failed ({num_fails}) through all retries "
                         f"({num_retries})")
                return None

            # Validation failed, retry
            line = on_fail(response)
            if line:
                LOG.debug(f"Speaking failure dialog: {line}")
                self.speak(line, wait=True, message=message, private=True)

            LOG.debug("Listen for another response")
            msg = message.reply('mycroft.mic.listen') or \
                Message('mycroft.mic.listen',
                        context={"skill_id": self.skill_id})
            self.bus.emit(msg)

    def __get_response(self, user="local"):
        """
        Helper to get a response from the user

        Arguments:
            user (str): user associated with response
        Returns:
            str: user's response or None on a timeout
        """
        event = Event()

        def converse(message):
            resp_user = get_message_user(message) or "local"
            if resp_user == user:
                utterances = message.data.get("utterances")
                converse.response = utterances[0] if utterances else None
                event.set()
                LOG.info(f"Got response: {converse.response}")
                return True
            LOG.debug(f"Ignoring input from: {resp_user}")
            return False

        # install a temporary conversation handler
        self.make_active()
        converse.response = None
        default_converse = self.converse
        self.converse = converse

        if not event.wait(self._get_response_timeout):
            LOG.warning("Timed out waiting for user response")
        self.converse = default_converse
        return converse.response



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import socket

try:
    import netifaces
    import requests
    from requests.exceptions import MissingSchema, InvalidSchema, InvalidURL, \
        ConnectionError
except ImportError:
    raise ImportError("netifaces or requests not available,"
                      " pip install neon-utils[network]")


def get_ip_address() -> str:
    """
    Returns the IPv4 address of the default interface (This is a public IP for server implementations)
    :return: IP Address
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(("8.8.8.8", 80))
    return s.getsockname()[0]


def get_adapter_info(interface: str = "default") -> dict:
    """
    Returns MAC and IP info for the specified gateway
    :param interface: Name of network interface to check
    :return: Dict of mac, IPv4 and IPv6 addresses
    """
    gateways = netifaces.gateways()
    if interface not in gateways:
        raise IndexError("Requested Interface Not Found")
    if netifaces.AF_INET not in gateways[interface]:
        raise IndexError("Requested Interface not connected!")
    device = gateways[interface][netifaces.AF_INET][1]
    mac = netifaces.ifaddresses(device)[netifaces.AF_LINK][0]['addr']
    ip4 = netifaces.ifaddresses(device)[netifaces.AF_INET][0]['addr']
    ip6 = netifaces.ifaddresses(device)[netifaces.AF_INET6][0]['addr']
    return {"ipv4": ip4, "ipv6": ip6, "mac": mac}


def check_url_response(url: str = "https://google.com") -> bool:
    """
    Checks if the passed url is accessible.
    :param url: URL to connect to (http/https schema expected)
    :returns: resp.ok if request is completed, False if ConnectionError is raised
    """
    if not isinstance(url, str):
        raise ValueError(f"{url} is not a str")
    try:
        resp = requests.get(url)
        return resp.ok
    except MissingSchema:
        return check_url_response(f"http://{url}")
    except InvalidSchema:
        raise ValueError(f"{url} is not a valid http url")
    except InvalidURL:
        raise ValueError(f"{url} is not a valid URL")
    except ConnectionError:
        # Offline Response
        return False
    except Exception as e:
        raise e


def check_online(valid_urls: tuple = ("https://google.com", "https://github.com")) -> bool:
    """
    Checks if device is online by pinging expected online remote URLs
    :param valid_urls: list of URL's to use
    :returns: True if remote sites can be reached, else False
    """
    if not isinstance(valid_urls, tuple):
        raise ValueError(f"Expected tuple, got {type(valid_urls)}")
    for url in valid_urls:
        try:
            if check_url_response(url):
                return True
        except ValueError:
            pass

    return False


def check_port_is_open(addr: str, port: int) -> bool:
    """
    Checks if the specified port at addr is open
    :param addr: IP or URL to query
    :param port: port to check
    :returns: True if the port is reachable, else False
    """
    test_connection = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    test_connection.settimeout(5)
    status = test_connection.connect_ex((addr, port))
    return status == 0




# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pendulum

from datetime import datetime
from typing import Optional, Union
from dateutil.tz import tzlocal
from geopy.exc import GeocoderServiceError
from geopy.geocoders import Nominatim
from timezonefinder import TimezoneFinder
from re import sub
from ovos_utils.log import LOG


# geocode.maps.co nominatim.openstreetmap.org
_NOMINATIM_DOMAIN = "nominatim.openstreetmap.org"


def set_nominatim_domain(domain: str):
    """
    Configure the Nominatim domain to use for location requests
    :param domain: Nominatim domain to use for location lookups
    """
    global _NOMINATIM_DOMAIN
    if domain.startswith("http"):
        domain = domain.split('/')[2]
    _NOMINATIM_DOMAIN = domain


def get_full_location(address: Union[str, tuple],
                      lang: Optional[str] = None) -> Optional[dict]:
    """
    Get full location details for the specified address in the specified lang
    :param address: string address or tuple (latitude, longitude)
    :param lang: optional language to format results (else system default)
    :returns: dict containing at minimum `place_id`, `lat`, `lon`, `address`
        None if service is not available
    """
    try:
        nominatim = Nominatim(user_agent="neon-ai", domain=_NOMINATIM_DOMAIN,
                              timeout=10)
        if isinstance(address, str):
            location = nominatim.geocode(address, addressdetails=True,
                                         language=lang)
        else:
            location = nominatim.reverse(address, addressdetails=True,
                                         language=lang)

        dict_location = location.raw
        dict_location['address']['country'] = sub(f'[0-9]', '',
                                                  dict_location['address'].
                                                  get('country'))
        return dict_location
    except GeocoderServiceError as e:
        LOG.error(e)
    except Exception as e:
        LOG.exception(e)
    return None


def get_coordinates(gps_loc: dict) -> (float, float):
    """
    Gets the latitude and longitude for the passed location
    :param gps_loc: dict of "city", "state", "country"
    :return: lat, lng float values
    """
    coordinates = Nominatim(user_agent="neon-ai", domain=_NOMINATIM_DOMAIN,
                            timeout=10)
    try:
        location = coordinates.geocode(gps_loc)
        LOG.debug(f"{location}")
        return location.latitude, location.longitude
    except GeocoderServiceError as e:
        LOG.error(e)
    except Exception as x:
        LOG.exception(x)
    return -1, -1


def get_location(lat, lng) -> (str, str, str, str):
    """
    Gets location name values for the passed coordinates.
    Note that some coordinates do not have a city, but may have a county.
    :param lat: latitude
    :param lng: longitude
    :return: city, county, state, country
    """
    try:
        address = Nominatim(user_agent="neon-ai", domain=_NOMINATIM_DOMAIN,
                            timeout=10)
    except GeocoderServiceError as e:
        LOG.error(e)
        return None
    except Exception as x:
        LOG.exception(x)
        return None
    location = address.reverse([lat, lng], language="en-US")
    LOG.debug(f"{location}")
    LOG.debug(f"{location.raw}")
    LOG.debug(f"{location.raw.get('address')}")
    city = location.raw.get('address').get('city') or \
        location.raw.get('address').get('town') or \
        location.raw.get('address').get('village') or \
        location.raw.get('address').get('hamlet')
    county = location.raw.get('address').get('county')
    state = location.raw.get('address').get('state')
    country = location.raw.get('address').get('country')
    return city, county, state, country


def get_timezone(lat, lng) -> (str, float):
    """
    Gets timezone information for the passed coordinates.
    Note that some coordinates do not have a city, but may have a county.
    :param lat: latitude
    :param lng: longitude
    :return: timezone name, offset from GMT
    """
    timezone = TimezoneFinder().timezone_at(lng=float(lng), lat=float(lat))
    offset = pendulum.from_timestamp(0, timezone).offset_hours
    return timezone, offset


def to_system_time(dt: datetime) -> datetime:
    """
    Converts a timezone aware or timezone naiive datetime object to a datetime
    object in the system tz
    :param dt: datetime object to convert
    :return: timezone aware datetime object that can be scheduled
    """
    tz = tzlocal()
    if dt.tzinfo:
        return dt.astimezone(tz)
    else:
        return dt.replace(tzinfo=tz).astimezone(tz)



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import nltk
import re
import os

from neon_utils.logger import LOG


def clean_quotes(raw_utt: str) -> str:
    """
    Method for stripping quotes from fully quoted strings in different languages
    :param raw_utt: Input string to be cleaned
    :return: string with all paired quote characters removed
    """
    if raw_utt is None:
        raise ValueError("Expected a string and got None")
    if not isinstance(raw_utt, str):
        raise TypeError(f"{raw_utt} is not a string!")

    leading_quotes = ['“', '"', '«', '「', "'", u'\u201d', u'\u00bb',
                      u'\u201e', u'u\xa0', u'\u00a0']
    trailing_quotes = ['“', '"', '»', '」', "'", u'\u201d', u'\u00bb',
                       u'\u201e', u'u\xa0', u'\u00a0']
    chars_to_remove = {*leading_quotes, *trailing_quotes}
    raw_utt = raw_utt.strip()
    utt = raw_utt
    trailing_punctuation = False
    if utt.endswith("."):
        trailing_punctuation = True
        utt = utt.rstrip(".")
    quotes_cleaned = False
    try:
        # Checks if utterance starts AND ends with some form of quotations and removes them accordingly
        while any((utt.startswith(lq) for lq in leading_quotes)) and \
                any((utt.endswith(tq) for tq in trailing_quotes)):
            quotes_cleaned = True
            removed_left, removed_right = False, False
            for c in chars_to_remove:
                if not removed_left and utt.startswith(c):
                    utt = utt[1:]
                    removed_left = True
                if not removed_right and utt.endswith(c):
                    utt = utt[:-1]
                    removed_right = True
        if quotes_cleaned:
            if trailing_punctuation:
                return f"{utt}."
            return utt
        else:
            return raw_utt
    except Exception as x:
        LOG.error(x)
        return raw_utt


def clean_filename(raw_name: str, to_lowercase: bool = False) -> str:
    """
    Cleans a filename of any invalid characters
    :param raw_name: input file basename to clean
    :param to_lowercase: cast string to lowercase
    :return: cleaned file basename
    """
    if raw_name is None:
        raise ValueError("Expected a string and got None")
    invalid_chars = ('/', '\\', '*', '~', ':', '"', '<', '>', '|', '?')
    name = raw_name
    for char in invalid_chars:
        name = name.replace(char, "_")
    if to_lowercase:
        name = name.lower()
    return name


def clean_transcription(raw_string: str) -> str:
    """
    Cleans up input transcriptions to replace any special characters with text
    :param raw_string: Input string to be cleaned
    :return: Cleaned string of alphas
    """
    if raw_string is None:
        raise ValueError("Expected a string and got None")
    parsed = raw_string.lower().replace('%', ' percent').replace('.', ' ')\
        .replace('?', '').replace('-', ' ').strip()
    # TODO: Cleanup to alpha string (replace all chars)
    return parsed


def get_phonemes(phrase: str) -> str:
    """
    Gets phonemes for the requested phrase
    :param phrase: String phrase for which to get phonemes
    :return: ARPAbet phonetic representation
        (https://en.wikipedia.org/wiki/ARPABET)
    """
    download_path = os.path.expanduser("~/.local/share/neon")
    if not os.path.isdir(download_path):
        os.makedirs(download_path)
    nltk.download('cmudict', download_dir=download_path)
    nltk.data.path.append(download_path)

    output = ''
    for word in phrase.split():
        phoenemes = nltk.corpus.cmudict.dict()[word.lower()][0]
        for phoeneme in phoenemes:
            output += str(re.sub('[0-9]', '', phoeneme) + ' ')
        output += '. '
    return output.rstrip()


def format_speak_tags(sentence: str, include_tags: bool = True) -> str:
    """
    Cleans up SSML tags for speech synthesis and ensures the phrase is wrapped
    in 'speak' tags and any excluded text is removed.
    :param sentence: Input sentence to be spoken
    :param include_tags: Flag to include <speak> tags in returned string
    :return: Cleaned sentence to pass to TTS
    """
    # Wrap sentence in speak tag if no tags present
    if "<speak>" not in sentence and "</speak>" not in sentence:
        to_speak = f"<speak>{sentence}</speak>"
    # Assume speak starts at the beginning of the sentence
    # if a closing speak tag is found
    elif "<speak>" not in sentence:
        to_speak = f"<speak>{sentence}"
    # Assume speak ends at the end of the sentence
    # if an opening speak tag is found
    elif "</speak>" not in sentence:
        to_speak = f"{sentence}</speak>"
    else:
        to_speak = sentence

    # Trim text outside of speak tags
    if not to_speak.startswith("<speak>"):
        to_speak = f"<speak>{to_speak.split('<speak>', 1)[1]}"

    if not to_speak.endswith("</speak>"):
        to_speak = f"{to_speak.split('</speak>', 1)[0]}</speak>"

    if to_speak == "<speak></speak>":
        return ""

    if include_tags:
        return to_speak
    else:
        return to_speak.lstrip("<speak>").rstrip("</speak>")


def normalize_string_to_speak(to_speak: str) -> str:
    """
    Normalizes spoken strings for TTS engines to handle
    :param to_speak: String to speak
    :return: string with any invalid characters removed and punctuation added
    """
    if to_speak is None:
        raise ValueError("Expected a string and got None")
    if not isinstance(to_speak, str):
        raise TypeError(f"{to_speak} is not a string!")

    valid_punctuation = ['.', '?', '!']
    if any(to_speak.endswith(x) for x in valid_punctuation):
        return to_speak
    return f"{to_speak}."


def transliteration(transcription: str, text: str, lang: str) -> str:
    '''
    Transliterates string from transcription provided by stt plugins
    Transliterates string if transcribed string length equals filename length
    :param transcription: input text to transliterate
    :param text: input text from audio filename
    :param lang: language of input audio -> ISO-639-1 code
    :return: transliterated or raw string
    '''
    transliterated = []
    translit_dict = {}
    if lang == 'pl':
        translit_dict = {'a': ['ą'], 'c': ['ć'], 'e': ['ę'], 'n': ['ń'],
                         'o': ['ó'], 's': ['ś'], 'z': ['ź', 'ż'], 'l': ['ł']}
    elif lang == 'fr':
        translit_dict = {'c': ['ç'], 'e': ['é', 'ê', 'è', 'ë'], 'a': ['â', 'à'],
                         'i': ['î', 'ì', 'ï'], 'o': ['ô', 'ò'],
                         'u': ['û', 'ù', 'ü']}
    elif lang == 'es':
        translit_dict = {'a': ['á'], 'i': ['í'], 'e': ['é'], 'n': ['ñ'],
                         'o': ['ó'], 'u': ['ú', 'ü']}
    elif lang == 'de':
        translit_dict = {'a': ['ä'], 's': ['ß'], 'o': ['ö'], 'u': ['ü']}
    transcription = re.sub("`|'|-", "", transcription)
    text = re.sub("`|'|-", "", text)
    if len(transcription.strip()) == len(text.strip()):
        for ind, letter in enumerate(transcription):
            if letter in translit_dict.keys():
                if letter != text[ind]:
                    for l in translit_dict[letter]:
                        if l == text[ind]:
                                transliterated.append(l)
                else:
                        transliterated.append(letter)
            else:
                    transliterated.append(letter)
        translit_str = ''.join(transliterated)
        if translit_str != '':
            return translit_str
        else:
            return text
    else:
        return text


def validate_email(email_addr: str) -> str:
    """
    Check if the passed string is a valid email address and try to parse it
    into one.
    :param email_addr: email address to validate
    :returns: Validated email address or an empty string
    """
    email_rx = r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,7}\b'
    if re.fullmatch(email_rx, email_addr):
        return email_addr

    if ' ' in email_addr:
        LOG.info(f"Removing whitespace from email address")
        email_addr = email_addr.replace(' ', '')
        if re.fullmatch(email_rx, email_addr):
            return email_addr

    return ""



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import io
import logging
import os
import sys
import shutil
import unittest

from os.path import basename, isdir
from time import time, sleep
from datetime import timedelta

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

ROOT_DIR = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
LOG_PATH = os.path.join(ROOT_DIR, "tests", "log_files")


class LogUtilTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        os.environ['XDG_CONFIG_HOME'] = '/tmp'
        os.makedirs(LOG_PATH, exist_ok=True)

    @classmethod
    def tearDownClass(cls) -> None:
        os.environ.pop('XDG_CONFIG_HOME')
        shutil.rmtree(LOG_PATH)

    def test_get_log_file(self):
        from neon_utils.log_utils import get_logger
        log = get_logger("test", LOG_PATH)
        log.level = logging.DEBUG
        test_msg = "This should be in test.log"
        log.debug(test_msg)
        with open(os.path.join(LOG_PATH, "test.log")) as log:
            contents = log.read()
        self.assertTrue(contents.endswith(f"{test_msg}\n"))

    def test_get_log_file_with_stdout(self):
        from neon_utils.log_utils import get_logger
        normal_stdout = sys.stdout
        captured_out = io.StringIO()
        sys.stdout = captured_out
        log = get_logger("test_stdout", LOG_PATH, True)
        log.level = logging.DEBUG
        test_msg = "This should be in test.log and stdout"
        log.debug(test_msg)
        with open(os.path.join(LOG_PATH, "test_stdout.log")) as log:
            contents = log.read()
        logged = captured_out.getvalue().strip("\n")
        sys.stdout = normal_stdout
        self.assertTrue(contents.endswith(f"{test_msg}\n"))
        self.assertEqual(logged, contents.split("\n")[-2])

    def test_get_log_no_file(self):
        from neon_utils.log_utils import get_logger
        logs = os.listdir(LOG_PATH)
        normal_stdout = sys.stdout
        captured_out = io.StringIO()
        sys.stdout = captured_out
        log = get_logger("terminal_only", "stdout")
        log.level = logging.DEBUG
        test_msg = "This should be in stdout ONLY"
        log.debug(test_msg)
        logged = captured_out.getvalue().strip("\n")
        sys.stdout = normal_stdout
        self.assertTrue(logged.endswith(test_msg))
        self.assertEqual(logs, os.listdir(LOG_PATH))

    def test_archive_logs_default(self):
        from neon_utils.log_utils import archive_logs
        os.makedirs(LOG_PATH, exist_ok=True)
        test_log = os.path.join(LOG_PATH, "to_backup.log")
        with open(test_log, "w+") as f:
            f.write("TEST LOG")
        archive_logs(LOG_PATH)
        for path in os.listdir(LOG_PATH):
            if os.path.isdir(os.path.join(LOG_PATH, path)):
                self.assertTrue(os.path.isfile(os.path.join(LOG_PATH, path, "to_backup.log")))
                return
        self.assertTrue(False)

    def test_archive_logs_specific(self):
        from neon_utils.log_utils import archive_logs
        os.makedirs(LOG_PATH, exist_ok=True)
        test_log = os.path.join(LOG_PATH, "to_backup.log")
        with open(test_log, "w+") as f:
            f.write("TEST LOG")
        archive_logs(LOG_PATH, "backup")
        path = os.path.join(LOG_PATH, "backup")
        self.assertTrue(os.path.isfile(os.path.join(LOG_PATH, path, "to_backup.log")))

    def test_remove_old_logs(self):
        from neon_utils.log_utils import archive_logs, remove_old_logs, init_log_for_module
        init_log_for_module(std_out=True)
        os.makedirs(LOG_PATH, exist_ok=True)
        test_log = os.path.join(LOG_PATH, "to_be_removed.log")
        with open(test_log, "w+") as f:
            f.write("TEST LOG")
        archive_logs(LOG_PATH)
        old_log_time = time()
        sleep(1)

        test_log = os.path.join(LOG_PATH, "to_be_retained.log")
        with open(test_log, "w+") as f:
            f.write("TEST LOG")
        archive_logs(LOG_PATH)

        remove_old_logs(LOG_PATH, timedelta(seconds=time() - old_log_time))
        self.assertEqual(len([p for p in os.listdir(LOG_PATH) if os.path.isdir(os.path.join(LOG_PATH, p))]), 1)
        for path in os.listdir(LOG_PATH):
            if os.path.isdir(os.path.join(LOG_PATH, path)):
                self.assertTrue(os.path.isfile(os.path.join(LOG_PATH, path, "to_be_retained.log")))
                return

    def test_get_log_file_for_module(self):
        from neon_utils.log_utils import get_log_file_for_module
        self.assertEqual("voice.log", os.path.basename(get_log_file_for_module("neon_speech_client")))
        self.assertEqual("voice.log", os.path.basename(get_log_file_for_module("neon_speech")))
        self.assertEqual("bus.log", os.path.basename(get_log_file_for_module(["python3", "-m",
                                                                              "mycroft.messagebus.service"])))
        self.assertEqual("bus.log", os.path.basename(get_log_file_for_module("neon_messagebus_service")))
        self.assertEqual("skills.log", os.path.basename(get_log_file_for_module("mycroft.skills")))
        self.assertEqual("skills.log", os.path.basename(get_log_file_for_module("neon_skills_service")))
        self.assertEqual("gui.log", os.path.basename(get_log_file_for_module("mycroft-gui-app")))
        self.assertEqual("display.log", os.path.basename(get_log_file_for_module("neon_gui_service")))
        self.assertEqual("display.log", os.path.basename(get_log_file_for_module("neon_core.gui")))

        self.assertEqual("extras.log", os.path.basename(get_log_file_for_module("NGI.gui")))
        self.assertEqual("extras.log", os.path.basename(get_log_file_for_module("nothing")))

    def test_init_log_for_module(self):
        from neon_utils.log_utils import init_log_for_module, ServiceLog, LOG
        init_log_for_module(ServiceLog.SPEECH)
        self.assertEqual(basename(LOG.base_path), ServiceLog.SPEECH.value)
        init_log_for_module(ServiceLog.AUDIO, True)
        self.assertEqual(LOG.base_path, "stdout")
        init_log_for_module(ServiceLog.AUDIO)
        self.assertEqual(basename(LOG.base_path), ServiceLog.AUDIO.value)

        from neon_utils.logger import LOG as neon_log
        from ovos_utils.log import LOG as ovos_log
        self.assertEqual(neon_log, ovos_log)
        self.assertEqual(neon_log, LOG)

    def test_get_log_dir(self):
        from neon_utils.log_utils import get_log_dir
        log_dir = get_log_dir()
        self.assertTrue(isdir(log_dir))

    def test_init_log(self):
        from neon_utils.log_utils import init_log
        config = {'log_level': 'DEBUG',
                  "logs": {
                      "name": "neon-utils"
                  }}
        log = init_log(config)
        self.assertEqual(log.name, "neon-utils")
        self.assertEqual(log.level, "DEBUG")
        from ovos_utils.log import LOG as OLOG
        from neon_utils.logger import LOG as NLOG
        self.assertEqual(OLOG, log)
        self.assertEqual(NLOG, log)

        config['log_level'] = 'INFO'
        config['logs']['name'] = 'test'
        config['logs']['level_overrides'] = {
            'error': ['filelock']
        }
        new_log = init_log(config)
        self.assertEqual(log, new_log)
        self.assertEqual(log.name, "test")
        self.assertEqual(log.level, "INFO")
        self.assertEqual(logging.getLogger('filelock').level, logging.ERROR)
        self.assertEqual(OLOG, log)
        self.assertEqual(NLOG, log)

    def test_get_log(self):
        from neon_utils.log_utils import get_log
        from ovos_utils.log import LOG as OLOG
        from neon_utils.log_utils import LOG as NLOG
        self.assertEqual(get_log(), OLOG)
        self.assertEqual(get_log(), NLOG)


if __name__ == '__main__':
    unittest.main()



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import logging

from datetime import datetime, timedelta
from enum import Enum
from os.path import isdir
from typing import Optional, Union
from ovos_utils.xdg_utils import xdg_data_home
from ovos_utils.log import LOG


_LOG = None


def get_log_dir(config: dict = None) -> str:
    """
    Get log directory from configuration or default path, create if not exists
    """
    from ovos_config.config import Configuration
    config = config or Configuration()
    log_dir = os.path.expanduser(config.get("log_dir") or
                                 os.path.join(xdg_data_home(), "neon", "logs"))
    if not isdir(log_dir):
        os.makedirs(log_dir, exist_ok=True)
    return log_dir


class ServiceLog(Enum):
    SPEECH = "voice.log"
    SKILLS = "skills.log"
    AUDIO = "audio.log"
    ENCLOSURE = "enclosure.log"
    BUS = "bus.log"
    GUI = "gui.log"
    DISPLAY = "display.log"
    SERVER = "server.log"
    CLIENT = "client.log"
    OTHER = "extras.log"


def remove_old_logs(log_dir: str = None,
                    history_to_retain: timedelta = timedelta(weeks=6)):
    """
    Removes archived logs older than the specified history timedelta
    Args:
        log_dir: Path to archived logs
        history_to_retain: Timedelta of history to retain
    """
    from shutil import rmtree
    log_dir = log_dir or get_log_dir()
    for archive in os.listdir(log_dir):
        archive_path = os.path.join(log_dir, archive)
        if not os.path.isdir(archive_path):
            continue
        if datetime.now() - datetime.fromtimestamp(os.path.getmtime(
                archive_path)) > history_to_retain:
            LOG.info(f"removing {archive}")
            rmtree(archive_path)


def archive_logs(log_dir: str = None, archive_dir: Optional[str] = None):
    """
    Archives the logs in the specified log_dir to log_dir/dir_name
    Args:
        log_dir: Path to log files to be archived
        archive_dir: Directory to archive logs to (defaults to formatted time)
    """
    from glob import glob
    from os.path import join, basename
    from os import makedirs
    from shutil import move
    log_dir = log_dir or get_log_dir()
    default_dirname = "logs--" + datetime.now().strftime("%Y-%m-%d--%H:%M:%S")
    archive_dir = join(log_dir, archive_dir or default_dirname)
    makedirs(archive_dir, exist_ok=True)
    for file in glob(join(log_dir, "*.log")):
        if basename(file) != "start.log":
            move(file, archive_dir)


def get_logger(log_name: str, log_dir: str = None,
               std_out: bool = False) -> logging.Logger:
    """
    Get a logger with the specified name and write to the specified log_dir
    and optionally std_out
    Args:
        log_name: Name of log (also used as log filename)
        log_dir: Directory to write log file to
        std_out: Flag to include logs in STDOUT

    Returns:
        Logger with the specified handlers
    """
    log_dir = log_dir or get_log_dir()
    LOG.init({"path": log_dir or "stdout"})
    LOG.name = log_name
    log = LOG.create_logger(log_name, std_out)
    return log


def get_log_file_for_module(module_name: Union[str, list]) -> str:
    """
    Gets the default log path for the requested module
    Args:
        module_name: Runnable argument passed to Popen
            (i.e. neon_speech_client, [python3, -m, mycroft.skills])

    Returns:
        Path to logfile
    """
    if isinstance(module_name, list):
        module_name = module_name[-1]
    if module_name.startswith("neon_speech"):
        log_name = "voice.log"
    elif module_name.startswith("neon_audio"):
        log_name = "audio.log"
    elif module_name.startswith("neon_enclosure"):
        log_name = "enclosure.log"
    elif any(x for x in ("neon_messagebus", "neon_core.messagebus",
                         "mycroft.messagebus") if module_name.startswith(x)):
        log_name = "bus.log"
    elif any(x for x in ("neon_skills", "neon_core.skills",
                         "mycroft.skills") if module_name.startswith(x)):
        log_name = "skills.log"
    elif any(x for x in ("neon_gui",
                         "neon_core.gui") if module_name.startswith(x)):
        log_name = "display.log"
    elif module_name == "neon_core_client":
        log_name = "client.log"
    elif module_name == "neon_core_server":
        log_name = "server.log"
    elif module_name == "mycroft-gui-app":
        log_name = "gui.log"
    else:
        log_name = "extras.log"

    return os.path.join(get_log_dir(), log_name)


def init_log_for_module(service: ServiceLog = ServiceLog.OTHER,
                        std_out: bool = False, max_bytes: int = 50000000,
                        backup_count: int = 3, level: str = logging.DEBUG):
    """
    Initialize `LOG` singleton for the specified service in this thread
    Args:
        service: service requesting a logger object
        std_out: if true, print logs to std_out instead of to files
        max_bytes: maximum size in bytes allowed for this log file
        backup_count: number of archived logs to save
        level: minimum log level to filter to
    """
    if not isdir(get_log_dir()):
        LOG.info(f"Creating log directory: {get_log_dir()}")
        os.makedirs(get_log_dir())
    log_file = "stdout" if std_out else os.path.join(get_log_dir(),
                                                     service.value)
    LOG.init({"path": log_file,
              "max_bytes": max_bytes,
              "backup_count": backup_count,
              "level": level})


def init_log(config: dict = None, log_name: str = None) -> type(LOG):
    """
    Initialize `LOG` with configuration params. Should be called once on module
    init.
    :param config: Configuration to apply to LOG
    :param log_name: Optional LOG.name override, else use Configuration or default
    :returns: LOG singleton
    """
    from ovos_config.config import Configuration
    _cfg = config or Configuration()
    _log_level = _cfg.get("log_level", "INFO")
    _logs_conf = _cfg.get("logs") or {}
    _logs_conf["level"] = _log_level
    LOG.name = log_name or _logs_conf.get("name") or "neon-utils"
    LOG.debug(f"Initializing logger with: {_logs_conf}")
    LOG.init(_logs_conf)  # read log level from config
    overrides = _logs_conf.get('level_overrides') or {}
    for log in overrides.get("error") or []:
        logging.getLogger(log).setLevel(logging.ERROR)
    for log in overrides.get("warning") or []:
        logging.getLogger(log).setLevel(logging.WARNING)
    for log in overrides.get("info") or []:
        logging.getLogger(log).setLevel(logging.INFO)
    for log in overrides.get("debug") or []:
        logging.getLogger(log).setLevel(logging.DEBUG)
    return LOG


def get_log() -> type(LOG):
    """
    Return an initialized `LOG` object
    """
    global _LOG
    if not _LOG:
        _LOG = init_log()
    return _LOG



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json
import os.path

from enum import Enum
from urllib.parse import urlparse
from neon_utils.logger import LOG


class CredentialEnvVar(Enum):
    """
    Mapping of service to well-known envvars
    """
    GITHUB = ("GH_TOKEN", "GITHUB_TOKEN")
    AWS_ID = ("AWS_ACCESS_KEY_ID",)
    AWS_SECRET = ("AWS_SECRET_ACCESS_KEY",)
    GOOGLE_CREDENTIALS = ("GOOGLE_APPLICATION_CREDENTIALS",)  # File ref
    WOLFRAM = ("WOLFRAM_APP_ID",)  # Neon Defined
    ALPHA_VANTAGE = ("ALPHA_VANTAGE_KEY",)  # Neon Defined
    OPEN_WEATHER_MAP = ("OWM_KEY",)  # Neon Defined


class CredentialNotFoundError(FileNotFoundError):
    """Credential not found in environ or requested file paths"""


def find_generic_keyfile(base_path: str, filename: str) -> str:
    """
    Locates a generic text keyfile
    Args:
        base_path: Base directory to check in addition to XDG directories
        (default ~/)
        filename: File basename to read
    Returns:
        str contents of located file
    """
    path_to_check = os.path.expanduser(base_path)
    paths_to_check = (path_to_check,
                      os.path.join(path_to_check, filename),
                      os.path.expanduser(f"~/.local/share/neon/{filename}"))
    for path in paths_to_check:
        if os.path.isfile(path):
            with open(path, "r") as f:
                credential = f.read().strip()
            return credential
    raise FileNotFoundError(f"No credentials found in default locations or "
                            f"path: {path_to_check}")


def find_environment_key(service: CredentialEnvVar):
    """
    Locate a credential in environment variables
    :param service: Credential to locate
    """
    for param in service.value:
        if os.getenv(param):
            return os.getenv(param)
    return None


def find_neon_git_token(base_path: str = "~/") -> str:
    """
    Searches environment variables and standard locations for a text file with
    a Github token.
    Args:
        base_path: Base directory to check in addition to XDG directories
        (default ~/)
    Returns:
        Github token string
    """
    env_token = find_environment_key(CredentialEnvVar.GITHUB)
    if env_token:
        return env_token
    try:
        return find_generic_keyfile(base_path, "token.txt")
    except FileNotFoundError:
        pass
    try:
        return find_generic_keyfile(base_path, "git_token.txt")
    except FileNotFoundError:
        pass

    raise CredentialNotFoundError("Could not locate github credentials")


def find_neon_aws_keys(base_path: str = "~/") -> dict:
    """
    Searches environment variables and standard locations for AWS credentials
    Args:
        base_path: Base directory to check in addition to XDG directories
        (default ~/)
    Returns:
        dict containing 'aws_access_key_id' and 'aws_secret_access_key'
    """
    env_aws_id = find_environment_key(CredentialEnvVar.AWS_ID)
    env_aws_key = find_environment_key(CredentialEnvVar.AWS_SECRET)
    if env_aws_id and env_aws_key:
        return {"aws_access_key_id": env_aws_id,
                "aws_secret_access_key": env_aws_key}

    try:
        csv_contents = find_generic_keyfile(base_path, "accessKeys.csv")
    except FileNotFoundError:
        csv_contents = None
    if csv_contents:
        try:
            aws_id, aws_key = csv_contents.split('\n')[1].split(',', 1)
            return {"aws_access_key_id": aws_id,
                    "aws_secret_access_key": aws_key}
        except Exception as e:
            LOG.error(e)

    try:
        json_contents = find_generic_keyfile(base_path, "aws.json")
    except FileNotFoundError:
        json_contents = None
    if json_contents:
        try:
            return json.loads(json_contents)
        except Exception as e:
            LOG.error(e)

    try:
        file_contents = find_generic_keyfile("~/.aws", "credentials")
    except FileNotFoundError:
        file_contents = None
    if file_contents:
        aws_id, aws_key = None, None
        for line in file_contents.split("\n"):
            if line.startswith("aws_access_key_id"):
                aws_id = line.split("=", 1)[1].strip()
            elif line.startswith("aws_secret_access_key"):
                aws_key = line.split("=", 1)[1].strip()
        if aws_id and aws_key:
            return {"aws_access_key_id": aws_id,
                    "aws_secret_access_key": aws_key}

    raise CredentialNotFoundError(f"No aws credentials found in default "
                                  f"locations or path: {base_path}")


def find_neon_google_keys(base_path: str = "~/") -> dict:
    """
    Locates google json credentials and returns the parsed
    credentials as a dict
    Args:
        base_path: Base directory to check in addition to XDG directories
        (default ~/)
    Returns:
        dict Google json credential
    """
    env_cred_file = find_environment_key(CredentialEnvVar.GOOGLE_CREDENTIALS)
    if env_cred_file and os.path.isfile(env_cred_file):
        try:
            with open(env_cred_file, 'r') as f:
                credential = json.load(f)
            return credential
        except Exception as e:
            LOG.error(f"Invalid google credential found at: {env_cred_file}")
            LOG.error(e)
    try:
        cred_str = find_generic_keyfile(base_path, "google.json")
        credential = json.loads(cred_str)
        return credential
    except Exception as e:
        LOG.error(e)
    raise CredentialNotFoundError(f"No google credentials found in default "
                                  f"locations or path: {base_path}")


def find_neon_wolfram_key(base_path: str = "~/") -> str:
    """
    Locates Wolfram|Alpha API key
    Args:
        base_path: Base directory to check in addition to XDG directories
         (default ~/)
    Returns:
        str Wolfram|Alpha API key
    """
    return find_environment_key(CredentialEnvVar.WOLFRAM) or \
        find_generic_keyfile(base_path, "wolfram.txt")


def find_neon_alpha_vantage_key(base_path: str = "~/") -> str:
    """
    Locates Alpha Vantage API key
    Args:
        base_path: Base directory to check in addition to XDG directories
         (default ~/)
    Returns:
        str Alpha Vantage API key
    """
    return find_environment_key(CredentialEnvVar.ALPHA_VANTAGE) or \
        find_generic_keyfile(base_path, "alpha_vantage.txt")


def find_neon_owm_key(base_path: str = "~/") -> str:
    """
    Locates Open Weather Map key
    Args:
        base_path: Base directory to check in addition to XDG directories
         (default ~/)
    Returns:
        str Open Weather Map API key
    """
    return find_environment_key(CredentialEnvVar.OPEN_WEATHER_MAP) or \
        find_generic_keyfile(base_path, "owm.txt")


def repo_is_neon(repo_url: str) -> bool:
    """
    Determines if the specified repository url is part of the NeonGeckoCom
    org on github
    Args:
        repo_url: string url to check
    Returns:
        True if the repository URL is known to be accessible using a neon
        auth key
    """
    url = urlparse(repo_url)
    if not url.scheme or not url.netloc:
        raise ValueError(f"{repo_url} is not a valid url")
    if any([x for x in ("github.com", "githubusercontent.com")
            if x in url.netloc]):
        try:
            author = url.path.split('/')[1]
        except IndexError:
            raise ValueError(f"{repo_url} is not a valid github url")
        if author.lower() == "neongeckocom":
            return True
        elif author.lower().startswith("neon"):
            # TODO: Get token and scrape org? DM
            LOG.info(f"Assuming repository uses Neon auth: {repo_url}")
            return True
    return False


def build_new_auth_config(key_path: str = "~/") -> dict:
    """
    Constructs a dict of authentication key data by locating credential files
        in the specified path
    :param key_path: path to locate key files
        (default locations checked in addition)
    :return: dict of located authentication keys
    """
    key_path = key_path or "~/"
    auth_config = {"github": {"token": find_neon_git_token},
                   "amazon": find_neon_aws_keys,
                   "wolfram": {"app_id": find_neon_wolfram_key},
                   "google": find_neon_google_keys,
                   "alpha_vantage": {"api_key": find_neon_alpha_vantage_key},
                   "owm": {"api_key": find_neon_owm_key}
                   }
    for cred in auth_config:
        try:
            if isinstance(auth_config[cred], dict):
                for real_cred in auth_config[cred]:
                    auth_config[cred][real_cred] = \
                        auth_config[cred][real_cred](key_path)
            else:
                auth_config[cred] = auth_config[cred](key_path)
        except FileNotFoundError:
            auth_config[cred] = dict()
            LOG.error(f"No credentials found for: {cred}")

    return auth_config



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json
import base64

# Limited maximum tcp packet size to 10 MB;
# Implied by the fact that TCP client aborts the connection once has its packet delivered to server
# thus preventing from sequential traversing
MAX_PACKET_SIZE = 10485760


def get_packet_data(socket, sequentially=False, batch_size=2048) -> bytes:
    """
        Gets all packet data by reading TCP socket stream sequentially
        :@param socket: TCP socket
        :@param sequentially: marker indicating whether received packet data should be read once or sequentially
        :@param batch_size: size of packet added through one sequence

        :@return bytes string representing the received data
    """
    if sequentially:
        fragments = []
        while True:
            chunk = socket.recv(batch_size)
            if not chunk:
                break
            fragments.append(chunk)
        data = b''.join(fragments)
    else:
        data = bytes(socket.recv(MAX_PACKET_SIZE))
    return data


def b64_to_dict(data: bytes, charset: str = "utf-8") -> dict:
    """
        Decodes base64-encoded message to python dictionary
        @param data: string bytes to decode
        @param charset: character set encoding to use (https://docs.python.org/3/library/codecs.html#standard-encodings)

        @return decoded dictionary
    """
    return eval(json.loads(base64.b64decode(data).decode(charset)))


def dict_to_b64(data: dict, charset: str = "utf-8") -> bytes:
    """
        Encodes python dictionary into base64 message
        @param data: python dictionary to encode
        @param charset: character set encoding to use (https://docs.python.org/3/library/codecs.html#standard-encodings)

        @return base64 encoded string
    """
    return base64.b64encode(json.dumps(str(data)).encode(charset))



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import json
import unittest
from os import remove
from os.path import join, dirname, isfile
from unittest.mock import patch


valid_config = {}
valid_headers = {}


class HanaUtilTests(unittest.TestCase):
    test_server = "https://hana.neonaialpha.com"
    test_path = join(dirname(__file__), "hana_test.json")

    def tearDown(self) -> None:
        global valid_config
        global valid_headers
        import neon_utils.hana_utils
        if isfile(self.test_path):
            remove(self.test_path)
        if neon_utils.hana_utils._client_config:
            valid_config = neon_utils.hana_utils._client_config
        if neon_utils.hana_utils._headers:
            valid_headers = neon_utils.hana_utils._headers
        neon_utils.hana_utils._client_config = {}
        neon_utils.hana_utils._headers = {}

    @patch("neon_utils.hana_utils._get_client_config_path")
    def test_request_backend(self, config_path):
        config_path.return_value = self.test_path

        # Use a valid config and skip extra auth
        import neon_utils.hana_utils
        neon_utils.hana_utils._client_config = valid_config
        neon_utils.hana_utils._headers = valid_headers
        from neon_utils.hana_utils import request_backend
        resp = request_backend("/neon/get_response",
                               {"lang_code": "en-us",
                                "utterance": "who are you",
                                "user_profile": {}}, self.test_server)
        self.assertEqual(resp['lang_code'], "en-us")
        self.assertIsInstance(resp['answer'], str)
        # TODO: Test invalid route, invalid request data

    @patch("neon_utils.hana_utils._get_client_config_path")
    def test_00_get_token(self, config_path):
        config_path.return_value = self.test_path
        from neon_utils.hana_utils import _get_token

        # Test valid request
        _get_token(self.test_server)
        from neon_utils.hana_utils import _client_config
        self.assertTrue(isfile(self.test_path))
        with open(self.test_path) as f:
            credentials_on_disk = json.load(f)
        self.assertEqual(credentials_on_disk, _client_config)
        # TODO: Test invalid request, rate-limited request

    @patch("neon_utils.hana_utils._get_client_config_path")
    @patch("neon_utils.hana_utils._get_token")
    def test_refresh_token(self, get_token, config_path):
        config_path.return_value = self.test_path
        import neon_utils.hana_utils

        def _write_token(*_, **__):
            with open(self.test_path, 'w+') as c:
                json.dump(valid_config, c)
            neon_utils.hana_utils._client_config = valid_config

        from neon_utils.hana_utils import _refresh_token
        get_token.side_effect = _write_token

        self.assertFalse(isfile(self.test_path))

        # Test valid request (auth + refresh)
        _refresh_token(self.test_server)
        get_token.assert_called_once()
        from neon_utils.hana_utils import _client_config
        self.assertTrue(isfile(self.test_path))
        with open(self.test_path) as f:
            credentials_on_disk = json.load(f)
        self.assertEqual(credentials_on_disk, _client_config)

        # Test refresh of existing token (no auth)
        _refresh_token(self.test_server)
        get_token.assert_called_once()
        with open(self.test_path) as f:
            new_credentials = json.load(f)
        self.assertNotEqual(credentials_on_disk, new_credentials)
        self.assertEqual(credentials_on_disk['client_id'],
                         new_credentials['client_id'])
        self.assertEqual(credentials_on_disk['username'],
                         new_credentials['username'])
        self.assertGreater(new_credentials['expiration'],
                           credentials_on_disk['expiration'])
        self.assertNotEqual(credentials_on_disk['access_token'],
                            new_credentials['access_token'])
        self.assertNotEqual(credentials_on_disk['refresh_token'],
                            new_credentials['refresh_token'])

    def test_config_path(self):
        from neon_utils.hana_utils import _get_client_config_path
        path_1 = _get_client_config_path("https://hana.neonaialpha.com")
        default = _get_client_config_path()
        self.assertNotEqual(path_1, default)
        self.assertEqual(dirname(path_1), dirname(default))

        # TODO: Test invalid refresh


if __name__ == '__main__':
    unittest.main()



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys
import os
import unittest

from copy import deepcopy
from os.path import join
from threading import Event
from unittest.mock import patch
from ovos_bus_client import Message

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))


class UserUtilTests(unittest.TestCase):
    def test_get_user_prefs(self):
        from neon_utils.user_utils import get_user_prefs

        test_user_1_profile = {"user": {"username": "test_user_1",
                                        "email": "test@neon.ai"}}
        test_user_2_profile = {"user": {"username": "test_user_2",
                                        "address": "new_key"}}
        test_message_1 = Message("test_message", {}, {
            "username": "test_user_1",
            "user_profiles": [test_user_1_profile,
                              test_user_2_profile]})
        test_message_2 = Message("test_message", {}, {
            "username": "test_user_2",
            "user_profiles": [test_user_1_profile,
                              test_user_2_profile]})

        user_1 = get_user_prefs(test_message_1)
        self.assertEqual(user_1["user"]["username"], "test_user_1")
        self.assertEqual(user_1["user"]["email"], "test@neon.ai")
        self.assertEqual(test_message_1.context['user_profiles'][0], user_1)

        user_2 = get_user_prefs(test_message_2)
        self.assertEqual(user_2["user"]["username"], "test_user_2")
        self.assertIn("address", user_2["user"])
        self.assertEqual(test_message_2.context['user_profiles'][1], user_2)

        missing_profile = get_user_prefs(Message("", {},
                                                 {"username": "test",
                                                  "user_profiles": [{}]}))
        self.assertEqual(missing_profile["user"]["username"], "test")
        missing_profile_2 = get_user_prefs(Message("", {},
                                                   {"username": "test2",
                                                    "user_profiles": [{}]}))
        self.assertEqual(missing_profile_2["user"]["username"], "test2")

        def wrapper(message, valid_dict):
            self.assertEqual(get_user_prefs(), valid_dict)

        wrapper(test_message_1, user_1)
        wrapper(test_message_2, user_2)

    @patch("ovos_config.config.Configuration")
    def test_get_default_user_config_from_mycroft_conf(self, config):
        from ovos_config.models import LocalConf
        # Patch configuration for test
        test_config_dir = os.path.join(os.path.dirname(__file__),
                                       "user_util_test_config")
        config.return_value = LocalConf(join(test_config_dir, "mycroft",
                                             "mycroft.conf"))
        import importlib
        from neon_utils import user_utils
        importlib.reload(user_utils)
        from neon_utils.user_utils import get_default_user_config
        user_config = get_default_user_config()
        self.assertFalse(os.path.isfile(os.path.join(test_config_dir,
                                                     "ngi_user_info.yml")))
        self.assertIsInstance(user_config, dict)
        self.assertEqual(user_config["location"],
                         {"lat": '38.971669',
                          "lng": '-95.23525',
                          "tz": 'America/Chicago',
                          "utc": '-6.0',
                          "city": 'Kirkland',
                          "state": 'Washington',
                          "country": "United States"})

    def test_update_user_profile(self):
        from neon_utils.user_utils import update_user_profile
        from ovos_utils.messagebus import FakeBus

        test_config_dir = os.path.join(os.path.dirname(__file__),
                                       "user_util_test_config")
        os.environ["NEON_CONFIG_PATH"] = test_config_dir
        os.environ["XDG_CONFIG_HOME"] = test_config_dir
        from neon_utils.user_utils import get_default_user_config
        user_config = get_default_user_config()
        user_config["user"]["username"] = "test_user"
        username = user_config["user"]["username"]
        self.assertIsInstance(user_config, dict)
        new_email = "developers@neon.ai"

        test_message = Message("test", {}, {"user_profiles": [user_config],
                                            "username": username})
        update_message = None
        updated = Event()
        bus = FakeBus()

        def _handle_update(message):
            nonlocal update_message
            update_message = message
            updated.set()
        bus.on("neon.profile_update", _handle_update)

        updated.clear()
        update_user_profile({"user": {"email": new_email}}, test_message, bus)
        self.assertEqual(
            test_message.context["user_profiles"][0]["user"]["email"],
            new_email)
        updated.wait(5)
        self.assertIsInstance(update_message, Message)
        self.assertEqual(test_message.context["user_profiles"][0],
                         update_message.data["profile"])
        self.assertEqual(update_message.data["profile"]["user"]["username"],
                         "test_user")

        valid_profile = deepcopy(update_message.data["profile"])
        updated.clear()
        update_user_profile({"invalid_key": {},
                             "user": {"invalid_key": "val"}},
                            test_message, bus)
        self.assertEqual(test_message.context["user_profiles"][0],
                         valid_profile)
        updated.wait(5)
        self.assertEqual(update_message.data["profile"], valid_profile)

    def test_update_default_config_on_profile_creation(self):
        test_config_dir = os.path.join(os.path.dirname(__file__),
                                       "user_util_test_config")
        os.environ["NEON_CONFIG_PATH"] = test_config_dir
        os.environ["XDG_CONFIG_HOME"] = test_config_dir
        import importlib
        from neon_utils.configuration_utils import get_neon_user_config
        from neon_utils import user_utils
        importlib.reload(user_utils)
        from neon_utils.user_utils import get_default_user_config
        user_config = get_default_user_config()
        self.assertFalse(os.path.isfile(os.path.join(test_config_dir,
                                                     "ngi_user_info.yml")))
        self.assertIsInstance(user_config, dict)

        from neon_utils.user_utils import update_user_profile, \
            apply_local_user_profile_updates
        from ovos_utils.messagebus import FakeBus
        bus = FakeBus()

        def on_profile_update(msg):
            # Mock behavior handled in IntentService
            updated_profile = msg.data.get("profile")
            apply_local_user_profile_updates(updated_profile,
                                             get_neon_user_config())

        bus.on("neon.profile_update", on_profile_update)
        local_user = get_neon_user_config()
        local_user['user']['username'] = 'local'
        message = Message("test", {}, {"username": local_user['user']['username'],
                                       "user_profiles": [local_user.content]})
        update_user_profile({"user": {"first_name": "Test",
                                      "last_name": "User"}}, message, bus)

        self.assertEqual(get_default_user_config()['user']['first_name'],
                         'Test')
        self.assertEqual(get_default_user_config()['user']['last_name'], 'User')

        os.remove(local_user.file_path)
        os.environ.pop("NEON_CONFIG_PATH")
        os.environ.pop("XDG_CONFIG_HOME")

    def test_apply_user_profile_updates(self):
        from neon_utils.user_utils import apply_local_user_profile_updates
        from neon_utils.configuration_utils import NGIConfig
        config_object = NGIConfig("test_config", os.path.dirname(__file__))
        config_object["test"] = {"updated": "",
                                 "unchanged": True
                                 }

        apply_local_user_profile_updates({"test": {"updated": "yes",
                                             "added": 1}}, config_object)
        self.assertEqual(dict(config_object.content),
                         {"test": {"updated": "yes",
                                   "unchanged": True,
                                   "added": 1}})
        self.assertFalse(config_object._pending_write)
        os.remove(os.path.join(config_object.path,
                               f".{config_object.name}.tmp"))
        os.remove(config_object.file_path)

    def test_get_default_user_config(self):
        from neon_utils.user_utils import get_default_user_config
        import neon_utils.user_utils
        neon_utils.user_utils._DEFAULT_USER_CONFIG = None
        user_config = get_default_user_config()
        self.assertEqual(set(user_config.keys()),
                         {'user', 'brands', 'location', 'units', 'speech',
                          'response_mode', 'privacy'})
        self.assertEqual(set(user_config['location'].keys()),
                         {'lat', 'lng', 'city', 'state', 'country', 'tz',
                          'utc'})


if __name__ == '__main__':
    unittest.main()
